import logging
from typing import Tuple, List, Union

import numpy as np
import utm
from sklearn.preprocessing import LabelEncoder
from tqdm.auto import tqdm

from cityseer.algos import data, checks
from cityseer.metrics import networks

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def dict_wgs_to_utm(data_dict: dict) -> dict:
    """
    Converts data dictionary `x` and `y` values from [WGS84](https://epsg.io/4326) `lng`, `lat` geographic coordinates
    to the local UTM projected coordinate system.

    Parameters
    ----------
    data_dict
        A dictionary representing distinct data points, where each `key` represents a `uid` and each value represents a nested dictionary with `x` and `y` key-value pairs.
        ```python
        example_data_dict = {
            'uid_01': {
                'x': 6000956.463188213,
                'y': 600693.4059810264
            },
            'uid_02': {
                'x': 6000753.336609659,
                'y': 600758.7916663144
            }
        }
        ```

    Returns
    -------
    dict
        Returns a copy of the source dictionary with the `x` and `y` values converted to the local UTM coordinate
        system.

    Notes
    ------
    ```python
    from cityseer.tools import mock
    from cityseer.metrics import layers

    # let's generate a mock data dictionary
    G_wgs = mock.mock_graph(wgs84_coords=True)
    # mock_data_dict takes on the same extents on the graph parameter
    data_dict_WGS = mock.mock_data_dict(G_wgs, random_seed=25)
    # the dictionary now contains wgs coordinates
    for i, (key, value) in enumerate(data_dict_WGS.items()):
        print(key, value)
        # prints:
        # 0 {'x': -0.09600470559254023, 'y': 51.592916036617794}
        # 1 {'x': -0.10621770551738155, 'y': 51.58888719412964}
        if i == 1:
            break
            
    # any data dictionary that follows this template can be passed to dict_wgs_to_utm()
    data_dict_UTM = layers.dict_wgs_to_utm(data_dict_WGS)
    # the coordinates have now been converted to UTM
    for i, (key, value) in enumerate(data_dict_UTM.items()):
        print(key, value)
        # prints:
        # 0 {'x': 701144.5207785056, 'y': 5719758.706109629}
        # 1 {'x': 700455.0000341447, 'y': 5719282.703221394}
        if i == 1:
            break
    ```
    """
    if not isinstance(data_dict, dict):
        raise TypeError('This method requires dictionary object.')

    logger.info('Converting data dictionary from WGS to UTM.')
    data_dict_copy = data_dict.copy()

    logger.info('Processing node x, y coordinates.')
    for k, v in tqdm(data_dict_copy.items(), disable=checks.quiet_mode):
        # x coordinate
        if 'x' not in v:
            raise AttributeError(f'Encountered node missing "x" coordinate attribute at data dictionary key {k}.')
        lng = v['x']
        # y coordinate
        if 'y' not in v:
            raise AttributeError(f'Encountered node missing "y" coordinate attribute at data dictionary key {k}.')
        lat = v['y']
        # check for unintentional use of conversion
        if abs(lng) > 180 or abs(lat) > 90:
            raise AttributeError('x, y coordinates exceed WGS bounds. Please check your coordinate system.')
        # be cognisant of parameter and return order
        # returns in easting, northing order
        easting, northing = utm.from_latlon(lat, lng)[:2]
        # write back to graph
        data_dict_copy[k]['x'] = easting
        data_dict_copy[k]['y'] = northing

    return data_dict_copy


def encode_categorical(classes: Union[list, tuple, np.ndarray]) -> Tuple[tuple, np.ndarray]:
    """
    Converts a list of land-use classes (or other categorical data) to an integer encoded version based on the unique
    elements.

    ::: warning Note
    It is generally not necessary to utilise this function directly. It will be called implicitly if calculating
    land-use metrics.
    :::

    Parameters
    ----------
    classes
        A `list`, `tuple` or `numpy` array of classes to be encoded.

    Returns
    -------
    tuple
        A `tuple` of unique class descriptors extracted from the input classes.
    np.ndarray
        A `numpy` array of the encoded classes. The value of the `int` encoding will correspond to the order of the
        `class_descriptors`.

    Notes
    -----
    ```python
    from cityseer.metrics import layers

    classes = ['cat', 'dog', 'cat', 'owl', 'dog']

    class_descriptors, class_encodings = layers.encode_categorical(classes)
    print(class_descriptors)  # prints: ('cat', 'dog', 'owl')
    print(list(class_encodings))  # prints: [0, 1, 0, 2, 1]
    ```
    """
    if not isinstance(classes, (list, tuple, np.ndarray)):
        raise TypeError('This method requires an iterable object.')

    # use sklearn's label encoder
    le = LabelEncoder()
    le.fit(classes)
    # map the int encodings to the respective classes
    classes_int = le.transform(classes)

    return tuple(le.classes_), classes_int


def data_map_from_dict(data_dict: dict) -> Tuple[tuple, np.ndarray]:
    """
    Converts a data dictionary into a `numpy` array for use by `DataLayer` classes.

    ::: warning Note
    It is generally not necessary to use this function directly. This function will be called implicitly when invoking [DataLayerFromDict](#class-datalayerfromdict)
    :::

    Parameters
    ----------
    data_dict
        A dictionary representing distinct data points, where each `key` represents a `uid` and each value represents a nested dictionary with `x` and `y` key-value pairs. The coordinates must be in a projected coordinate system matching that of the [`NetworkLayer`](/metrics/networks#class-networklayer) to which the data will be assigned.

        ```python
        example_data_dict = {
            'uid_01': {
                'x': 6000956.463188213,
                'y': 600693.4059810264
            },
            'uid_02': {
                'x': 6000753.336609659,
                'y': 600758.7916663144
            }
        }
        ```

    Returns
    -------
    tuple
        A tuple of data `uids` corresponding to the data point identifiers in the source `data_dict`.
    np.ndarray
        A 2d numpy array representing the data points. The indices of the second dimension correspond as follows:

        | idx | property |
        |-----|:----------|
        | 0   | `x` coordinate |
        | 1   | `y` coordinate |
        | 2   | assigned network index - nearest |
        | 3   | assigned network index - next-nearest |

        The arrays at indices `2` and `3` will be initialised with `np.nan`. These will be populated when the [DataLayer.assign_to_network](#datalayer-assign-to-network) method is invoked.
    """
    if not isinstance(data_dict, dict):
        raise TypeError('This method requires dictionary object.')

    data_uids = []
    data_map = np.full((len(data_dict), 4), np.nan)
    for i, (k, v) in enumerate(data_dict.items()):
        # set key to data labels
        data_uids.append(k)
        # DATA MAP INDEX POSITION 0 = x coordinate
        if 'x' not in v:
            raise AttributeError(f'Encountered entry missing "x" coordinate attribute at index {i}.')
        data_map[i][0] = v['x']
        # DATA MAP INDEX POSITION 1 = y coordinate
        if 'y' not in v:
            raise AttributeError(f'Encountered entry missing "y" coordinate attribute at index {i}.')
        data_map[i][1] = v['y']

    return tuple(data_uids), data_map


class DataLayer:
    """
    Categorical data, such as land-use classifications and numerical data, can be assigned to the network as a
    [`DataLayer`](/metrics/layers#class-datalayer). A `DataLayer` represents the spatial locations of data points, and
    can be used to calculate various mixed-use, land-use accessibility, and statistical measures. Importantly, these
    measures are computed directly over the street network and offer distance-weighted variants; the combination of
    which, makes them more contextually sensitive than methods otherwise based on simpler crow-flies aggregation
    methods.

    The coordinates of data points should correspond as precisely as possible to the location of the feature in space;
    or, in the case of buildings, should ideally correspond to the location of the building entrance.

    Note that in many cases, the [`DataLayerFromDict`](#class-#datalayerfromdict) class will provide a more convenient
    alternative for instantiating this class.
    """
    def __init__(self,
                 data_uids: Union[list, tuple],
                 data_map: np.ndarray):
        """


        Parameters
        ----------
        data_uids
            A `list` or `tuple` of data identifiers corresponding to each data point. This list must be in the same
            order and of the same length as the `data_map`.
        data_map
            A 2d `numpy` array representing the data points. The length of the first dimension should match that of the
            `data_uids`. The indices of the second dimension correspond as follows:

            | idx | property |
            |-----|:----------|
            | 0 | `x` coordinate |
            | 1 | `y` coordinate |
            | 2 | assigned network index - nearest |
            | 3 | assigned network index - next-nearest |

            The arrays at indices `2` and `3` will be populated when the
            [DataLayer.assign_to_network](#datalayer-assign-to-network) method is invoked.

        Returns
        -------
        DataLayer
            Returns a `DataLayer`.

        Properties
        ----------
        """

        self._uids = data_uids  # original labels / indices for each data point
        self._data = data_map  # data map per above
        self._Network = None

        # checks
        checks.check_data_map(self._data, check_assigned=False)

        if len(self._uids) != len(self._data):
            raise ValueError('The number of data labels does not match the number of data points.')

    @property
    def uids(self):
        """Unique ids corresponding to each location in the data_map."""
        return self._uids

    @property
    def data_x_arr(self):
        return self._data[:, 0]

    @property
    def data_y_arr(self):
        return self._data[:, 1]

    @property
    def Network(self):
        return self._Network

    def assign_to_network(self,
                          Network_Layer: networks.NetworkLayer,
                          max_dist: Union[int, float]):
        """
        Once created, a [`DataLayer`](#class-datalayer) should be assigned to a [`NetworkLayer`](/metrics/networks#class-networklayer). The
        `NetworkLayer` provides the backbone for the localised spatial aggregation of data points over the street
        network. The measures will be computed over the same distance thresholds as used for the `NetworkLayer`.

        The data points will be assigned to the two closest network nodes — one in either direction — based on the
        closest adjacent street edge. This enables a dynamic spatial aggregation method that more accurately describes
        distances over the network to data points, relative to the direction of approach.

        Parameters
        ----------
        Network_Layer
            A [`NetworkLayer`](/metrics/networks#class-networklayer).
        max_dist
            The maximum distance to consider when assigning respective data points to the nearest adjacent network
            nodes.

        Notes
        -----
        ::: tip Hint
        The `max_dist` parameter should not be set too small. There are two steps in the assignment process: the first,
        identifies the closest street node; the second, sets-out from this node and attempts to wind around the data
        point — akin to circling the block. It will then review the discovered graph edges from which it is able to
        identify the closest adjacent street-front. The `max_dist` parameter sets a crow-flies distance limit on how far
        the algorithm will search in its attempts to encircle the data point. If the `max_dist` is too small, then the
        algorithm is potentially hampered from finding a starting node; or, if a node is found, may have to terminate
        exploration prematurely because it can't travel far enough away from the data point to explore the surrounding
        network. If too many data points are not being successfully assigned to the correct street edges, then this
        distance should be increased. Conversely, if most of the data points are satisfactorily assigned, then it may be
        possible to decrease this threshold. A distance of $\approx 400m$ provides a good starting point.
        :::

        ::: warning Note
        The precision of assignment improves on decomposed networks (see
        [graphs.nX_decompose](/tools/graphs.html#nx-decompose)), which offers the additional benefit of a more granular
        representation of variations in metrics along street-fronts.
        :::

        ![Example assignment of data to a network](../.vitepress/plots/images/assignment.png)
        _Example assignment on a non-decomposed graph._

        ![Example assignment of data to a network](../.vitepress/plots/images/assignment_decomposed.png)
        _Assignment of data to network nodes becomes more contextually precise on decomposed graphs._
        """
        self._Network = Network_Layer
        data.assign_to_network(self._data,
                               self.Network._node_data,
                               self.Network._edge_data,
                               self.Network._node_edge_map,
                               max_dist,
                               suppress_progress=checks.quiet_mode)

    def compute_aggregated(self,
                           landuse_labels: Union[list, tuple, np.ndarray] = None,
                           mixed_use_keys: Union[list, tuple] = None,
                           accessibility_keys: Union[list, tuple] = None,
                           cl_disparity_wt_matrix: Union[list, tuple, np.ndarray] = None,
                           qs: Union[list, tuple, np.ndarray] = None,
                           stats_keys: Union[list, tuple] = None,
                           stats_data_arrs: Union[List[Union[list, tuple, np.ndarray]],
                                                  Tuple[Union[list, tuple, np.ndarray]],
                                                  np.ndarray] = None,
                           angular: bool = False):
        """
        This method wraps the underlying `numba` optimised functions for aggregating and computing various mixed-use,
        land-use accessibility, and statistical measures. These are computed simultaneously for any required
        combinations of measures (and distances), which can have significant speed implications. Situations requiring
        only a single measure can instead make use of the simplified [`DataLayer.hill_diversity`](#datalayer-hill-diversity),
        [`DataLayer.hill_branch_wt_diversity`](#datalayer-hill-branch-wt-diversity),
        [`DataLayer.compute_accessibilities`](#datalayer-compute-accessibilities),
        [`DataLayer.compute_stats_single`](#datalayer-compute-stats-single), and
        [`DataLayer.compute_stats_multiple`](#datalayer-compute-stats-multiple) methods.

        The data is aggregated and computed over the street network relative to the `Network Layer` nodes, with the
        implication that mixed-use, accessibility, and statistical aggregations are generated from the same locations as
        for centrality computations, which can therefore be correlated or otherwise compared. The outputs of the
        calculations are written to the corresponding node indices in the same `NetworkLayer.metrics` dictionary used
        for centrality methods, and will be categorised by the respective keys and parameters.

        For example, if `hill` and `shannon` mixed-use keys; `shops` and `factories` accessibility keys; and a
        `valuations` stats keys are computed on a `Network Layer` instantiated with $800m$ and $1600m$ distance
        thresholds, then the dictionary would assume the following structure:

        ```python
        NetworkLayer.metrics = {
            'mixed_uses': {
                # note that hill measures have q keys
                'hill': {
                    # here, q=0
                    0: {
                        800: [...],
                        1600: [...]
                    },
                    # here, q=1
                    1: {
                        800: [...],
                        1600: [...]
                    }
                },
                # non-hill measures do not have q keys
                'shannon': {
                    800: [...],
                    1600: [...]
                }
            },
            'accessibility': {
                # accessibility keys are computed in both weighted and unweighted forms
                'weighted': {
                    'shops': {
                        800: [...],
                        1600: [...]
                    },
                    'factories': {
                        800: [...],
                        1600: [...]
                    }
                },
                'non_weighted': {
                    'shops': {
                        800: [...],
                        1600: [...]
                    },
                    'factories': {
                        800: [...],
                        1600: [...]
                    }
                }
            },
            'stats': {
                # stats grouped by each stats key
                'valuations': {
                    # each stat will have the following key-value pairs
                    'max': {
                        800: [...],
                        1600: [...]
                    },
                    'min': {
                        800: [...],
                        1600: [...]
                    },
                    'sum': {
                        800: [...],
                        1600: [...]
                    },
                    'sum_weighted': {
                        800: [...],
                        1600: [...]
                    },
                    'mean': {
                        800: [...],
                        1600: [...]
                    },
                    'mean_weighted': {
                        800: [...],
                        1600: [...]
                    },
                    'variance': {
                        800: [...],
                        1600: [...]
                    },
                    'variance_weighted': {
                        800: [...],
                        1600: [...]
                    }
                }
            }
        }
        ```

        Parameters
        ----------
        landuse_labels
            A set of land-use labels corresponding to the length and order of the data points. The labels should
            correspond to descriptors from the land-use schema, such as "retail" or "commercial". This parameter is only
            required if computing mixed-uses or land-use accessibilities, by default None.
        mixed_use_keys
            An optional list of strings describing which mixed-use metrics to compute, containing any combination of
            `key` values from the following table, by default None. See **Notes** for additional information.
        accessibility_keys
            An optional `list` or `tuple` of land-use classifications for which to calculate accessibilities. The keys
            should be selected from the same land-use schema used for the `landuse_labels` parameter, e.g. "retail". The
            calculations will be performed in both `weighted` and `non_weighted` variants, by default None.
        cl_disparity_wt_matrix
            A pairwise `NxN` disparity matrix numerically describing the degree of disparity between any pair of
            distinct land-uses. This parameter is only required if computing mixed-uses using `hill_pairwise_disparity`
            or `raos_pairwise_disparity`.  The number and order of land-uses should match those implicitly generated by
            [`encode_categorical`](#encode-categorical), by default None.
        qs
            The values of `q` for which to compute Hill diversity. This parameter is only required if computing one of
            the Hill diversity mixed-use measures, by default None.
        stats_keys
            A `list` or `tuple` of keys corresponding to the number of nested arrays passed to the `stats_data_arrs`
            parameter. The computed stats will be saved to the `N.metrics` dictionary using these keys. This parameter
            is only required if computing stats for a `stats_data_arrs` parameter, by default None
        stats_data_arrs
            A 2d `list`, `tuple` or `numpy` array of numerical data, where the first dimension corresponds to the number
            of keys in the `stats_keys` parameter and the second dimension corresponds to number of data points in the
            `DataLayer`, by default None. e.g:
            ```python
            # for a DataLayer containg 5 data points
            stats_keys = ['valuations', 'floors', 'occupants']
            stats_data_arrs = [
                [50000, 60000, 55000, 42000, 46000],  # valuations
                [3, 3, 2, 3, 5],  # floors
                [420, 300, 220, 250, 600]  # occupants
            ]
            ```
        angular
            Whether to use a simplest-path heuristic in-lieu of a shortest-path heuristic when calculating aggregations
            and distances, by default False

        Notes
        -----
        | key | formula | notes |
        |-----|:-------:|-------|
        | hill | $\scriptstyle\big(\sum_{i}^{S}p_{i}^q\big)^{1/(1-q)}\ q\geq0,\ q\neq1 \\ \scriptstyle lim_{q\to1}\
        exp\big(-\sum_{i}^{S}\ p_{i}\ log\ p_{i}\big)$ | Hill diversity: this is the preferred form of diversity
        metric because it adheres to the replication principle and uses units of effective species instead of measures
        of information or uncertainty. The `q` parameter controls the degree of emphasis on the _richness_ of species as
        opposed to the _balance_ of species. Over-emphasis on balance can be misleading in an urban context, for which
        reason research finds support for using `q=0`: this reduces to a simple count of distinct land-uses.|
        | hill_branch_wt | $\scriptstyle\big[\sum_{i}^{S}d_{i}\big(\frac{p_{i}}{\bar{T}}\big)^{q} \big]^{1/(1-q)} \\
        \scriptstyle\bar{T} = \sum_{i}^{S}d_{i}p_{i}$ | This is a distance-weighted variant of Hill Diversity based
        on the distances from the point of computation to the nearest example of a particular land-use. It therefore
        gives a locally representative indication of the intensity of mixed-uses. $d_{i}$ is a negative exponential
        function where $-\beta$ controls the strength of the decay. ($-\beta$ is provided by the `Network Layer`, see
        [`distance_from_beta`](/metrics/networks#distance-from-beta).)|
        | hill_pairwise_wt | $\scriptstyle\big[ \sum_{i}^{S} \sum_{j\neq{i}}^{S} d_{ij} \big(  \frac{p_{i} p_{j}}{Q}
        \big)^{q} \big]^{1/(1-q)} \\ \scriptstyle Q = \sum_{i}^{S} \sum_{j\neq{i}}^{S} d_{ij} p_{i} p_{j}$ | This is a
        pairwise-distance-weighted variant of Hill Diversity based on the respective distances between the closest
        examples of the pairwise distinct land-use combinations as routed through the point of computation.
        $d_{ij}$ represents a negative exponential function where $-\beta$ controls the strength of the decay.
        ($-\beta$ is provided by the `Network Layer`, see
        [`distance_from_beta`](/metrics/networks#distance-from-beta).)|
        | hill_pairwise_disparity | $\scriptstyle\big[ \sum_{i}^{S} \sum_{j\neq{i}}^{S} w_{ij} \big(  \frac{p_{i}
        p_{j}}{Q} \big)^{q} \big]^{1/(1-q)} \\ \scriptstyle Q = \sum_{i}^{S} \sum_{j\neq{i}}^{S} w_{ij} p_{i}
        p_{j}$ | This is a disparity-weighted variant of Hill Diversity based on the pairwise disparities between
        land-uses. This variant requires the use of a disparity matrix provided through the `cl_disparity_wt_matrix`
        parameter.|
        | shannon | $\scriptstyle -\sum_{i}^{S}\ p_{i}\ log\ p_{i}$ | Shannon diversity (or_information entropy_) is
        one of the classic diversity indices. Note that it is preferable to use Hill Diversity with `q=1`, which is
        effectively a transformation of Shannon diversity into units of effective species.|
        | gini_simpson | $\scriptstyle 1 - \sum_{i}^{S} p_{i}^2$ | Gini-Simpson is another classic diversity index.
        It can behave problematically because it does not adhere to the replication principle and places emphasis on the
        balance of species, which can be counter-productive for purposes of measuring mixed-uses. Note that where an
        emphasis on balance is desired, it is preferable to use Hill Diversity with `q=2`, which is effectively a
        transformation of Gini-Simpson diversity into units of effective species.|
        | raos_pairwise_disparity | $\scriptstyle \sum_{i}^{S} \sum_{j \neq{i}}^{S} d_{ij} p_{i} p_{j}$ | Rao diversity
        is a pairwise disparity measure and requires the use of a disparity matrix provided through the
        `cl_disparity_wt_matrix` parameter. It suffers from the same issues as Gini-Simpson. It is preferable to use
        disparity weighted Hill diversity with `q=2`.|

        ::: tip Hint
        The available choices of land-use diversity measures may seem overwhelming. `hill_branch_wt` paired with `q=0`
        is generally the best choice for granular landuse data, or else `q=1` or `q=2` for increasingly crude landuse
        classifications schemas.
        :::

        A worked example:
        ```python
        from cityseer.metrics import networks, layers
        from cityseer.tools import mock, graphs

        # prepare a mock graph
        G = mock.mock_graph()
        G = graphs.nX_simple_geoms(G)

        # generate the network layer
        N = networks.NetworkLayerFromNX(G, distances=[200, 400, 800, 1600])

        # prepare a mock data dictionary
        data_dict = mock.mock_data_dict(G, random_seed=25)
        # prepare some mock land-use classifications
        landuses = mock.mock_categorical_data(len(data_dict), random_seed=25)
        # let's also prepare some numerical data
        stats_data = mock.mock_numerical_data(len(data_dict), num_arrs=1, random_seed=25)

        # generate a data layer
        L = layers.DataLayerFromDict(data_dict)
        # assign to the network
        L.assign_to_network(N, max_dist=500)
        # compute some metrics
        L.compute_aggregated(landuse_labels=landuses,
                            mixed_use_keys=['hill'],
                            qs=[0, 1],
                            accessibility_keys=['c', 'd', 'e'],
                            stats_keys=['mock_stat'],
                            stats_data_arrs=stats_data)
        # note that the above measures can be run individually using simplified interfaces, e.g.
        # L.hill_diversity(landuses, [0])
        # L.compute_accessibilities(landuses, ['a', 'b'])
        L.compute_stats_single('mock_stat', stats_data[0])  # this method requires a 1d array

        # let's prepare some keys for accessing the computational outputs
        # distance idx: any of the distances with which the NetworkLayer was initialised
        distance_idx = 200
        # q index: any of the invoked q parameters
        q_idx = 0
        # a node idx
        node_idx = 0

        # the data is available at N.metrics
        print(N.metrics['mixed_uses']['hill'][q_idx][distance_idx][node_idx])
        # prints: 4.0
        print(N.metrics['accessibility']['weighted']['d'][distance_idx][node_idx])
        # prints: 0.019168843947614676
        print(N.metrics['accessibility']['non_weighted']['d'][distance_idx][node_idx])
        # prints: 1.0
        print(N.metrics['stats']['mock_stat']['mean_weighted'][distance_idx][node_idx])
        # prints: 71297.82967202332
        ```

        Note that the data can also be unpacked to a dictionary using [`NetworkLayer.metrics_to_dict`](/metrics/networks#networklayer-metrics-to-dict), or transposed to a `networkX` graph using [`NetworkLayer.to_networkX`](/metrics/networks#networklayer-to-networkx).

        ::: danger Caution
        Be cognisant that mixed-use and land-use accessibility measures are sensitive to the classification schema that has been used. Meaningful comparisons from one location to another are only possible where the same schemas have been applied.
        :::
        """
        if self.Network is None:
            raise ValueError('Assign this data layer to a network prior to computing mixed-uses or accessibilities.')

        mixed_uses_options = ['hill',
                              'hill_branch_wt',
                              'hill_pairwise_wt',
                              'hill_pairwise_disparity',
                              'shannon',
                              'gini_simpson',
                              'raos_pairwise_disparity']

        if (stats_keys is not None and stats_data_arrs is None) \
                or (stats_keys is None and stats_data_arrs is not None) \
                or (stats_keys is not None and stats_data_arrs is not None and
                    len(stats_data_arrs) != len(stats_keys)):
            raise ValueError('An equal number of stats labels and stats data arrays is required.')

        if stats_data_arrs is None:
            stats_data_arrs = np.full((0, 0), np.nan)
        elif not isinstance(stats_data_arrs, (list, tuple, np.ndarray)):
            raise ValueError('Stats data must be in the form of a list, tuple, or numpy array.')
        else:
            stats_data_arrs = np.array(stats_data_arrs)
            if stats_data_arrs.ndim == 1:
                stats_data_arrs = np.array([stats_data_arrs])
            if stats_data_arrs.shape[1] != len(self._data):
                raise ValueError('The length of all data arrays must match the number of data points.')

        if landuse_labels is None:
            if mixed_use_keys or accessibility_keys:
                raise ValueError(
                    'Computation of mixed use or accessibility measures requires the landuse_labels parameter.')

            landuse_classes = ()
            landuse_encodings = ()
            qs = ()
            mu_hill_keys = ()
            mu_other_keys = ()
            acc_keys = ()
            cl_disparity_wt_matrix = [[]]  # (()) causes error because numpy conversion creates single dimension array

        # remember, most checks on parameter integrity occur in underlying method
        # so, don't duplicate here
        else:
            if len(landuse_labels) != len(self._data):
                raise ValueError('The number of landuse labels should match the number of data points.')

            # get the landuse encodings
            landuse_classes, landuse_encodings = encode_categorical(landuse_labels)

            # if necessary, check the disparity matrix
            if cl_disparity_wt_matrix is None:
                cl_disparity_wt_matrix = np.full((0, 0), np.nan)
            elif not isinstance(cl_disparity_wt_matrix, (list, tuple, np.ndarray)) or \
                    cl_disparity_wt_matrix.ndim != 2 or \
                    cl_disparity_wt_matrix.shape[0] != cl_disparity_wt_matrix.shape[1] or \
                    len(cl_disparity_wt_matrix) != len(landuse_classes):
                raise TypeError(
                    'Disparity weights must be a square pairwise NxN matrix in list, tuple, or numpy.ndarray form. '
                    'The number of edge-wise elements should match the number of unique class labels.')

            # warn if no qs provided
            if qs is None:
                qs = ()
            if isinstance(qs, (int, float)):
                qs = (qs)
            if not isinstance(qs, (list, tuple, np.ndarray)):
                raise TypeError('Please provide a float, list, tuple, or numpy.ndarray of q values.')

            # extrapolate the requested mixed use measures
            mu_hill_keys = []
            mu_other_keys = []
            if mixed_use_keys is not None:
                for mu in mixed_use_keys:
                    if mu not in mixed_uses_options:
                        raise ValueError(
                            f'Invalid mixed-use option: {mu}. Must be one of {", ".join(mixed_uses_options)}.')
                    idx = mixed_uses_options.index(mu)
                    if idx < 4:
                        mu_hill_keys.append(idx)
                    else:
                        mu_other_keys.append(idx - 4)
                if not checks.quiet_mode:
                    logger.info(f'Computing mixed-use measures: {", ".join(mixed_use_keys)}')

            acc_keys = []
            if accessibility_keys is not None:
                for ac_label in accessibility_keys:
                    if ac_label not in landuse_classes:
                        logger.warning(f'No instances of accessibility label: {ac_label} present in the data.')
                    else:
                        acc_keys.append(landuse_classes.index(ac_label))
                if not checks.quiet_mode:
                    logger.info(f'Computing land-use accessibility for: {", ".join(accessibility_keys)}')

        if stats_keys is not None and not checks.quiet_mode:
            logger.info(f'Computing stats for: {", ".join(stats_keys)}')

        # call the underlying method
        mixed_use_hill_data, mixed_use_other_data, \
        accessibility_data, accessibility_data_wt, \
        stats_sum, stats_sum_wt, \
        stats_mean, stats_mean_wt, \
        stats_variance, stats_variance_wt, \
        stats_max, stats_min = data.local_aggregator(self.Network._node_data,
                                                     self.Network._edge_data,
                                                     self.Network._node_edge_map,
                                                     self._data,
                                                     distances=np.array(self.Network.distances),
                                                     betas=np.array(self.Network.betas),
                                                     landuse_encodings=np.array(landuse_encodings),
                                                     qs=np.array(qs),
                                                     mixed_use_hill_keys=np.array(mu_hill_keys),
                                                     mixed_use_other_keys=np.array(mu_other_keys),
                                                     accessibility_keys=np.array(acc_keys),
                                                     cl_disparity_wt_matrix=np.array(cl_disparity_wt_matrix),
                                                     numerical_arrays=stats_data_arrs,
                                                     angular=angular,
                                                     suppress_progress=checks.quiet_mode)

        # write the results to the Network's metrics dict
        # keys will check for pre-existing, whereas qs and distance keys will overwrite
        # unpack mixed use hill
        for mu_h_idx, mu_h_key in enumerate(mu_hill_keys):
            mu_h_label = mixed_uses_options[mu_h_key]
            if mu_h_label not in self.Network.metrics['mixed_uses']:
                self.Network.metrics['mixed_uses'][mu_h_label] = {}
            for q_idx, q_key in enumerate(qs):
                self.Network.metrics['mixed_uses'][mu_h_label][q_key] = {}
                for d_idx, d_key in enumerate(self.Network.distances):
                    self.Network.metrics['mixed_uses'][mu_h_label][q_key][d_key] = \
                        mixed_use_hill_data[mu_h_idx][q_idx][d_idx]

        # unpack mixed use other
        for mu_o_idx, mu_o_key in enumerate(mu_other_keys):
            mu_o_label = mixed_uses_options[mu_o_key + 4]
            if mu_o_label not in self.Network.metrics['mixed_uses']:
                self.Network.metrics['mixed_uses'][mu_o_label] = {}
            # no qs
            for d_idx, d_key in enumerate(self.Network.distances):
                self.Network.metrics['mixed_uses'][mu_o_label][d_key] = mixed_use_other_data[mu_o_idx][d_idx]

        # unpack accessibility data
        for ac_idx, ac_code in enumerate(acc_keys):
            ac_label = landuse_classes[ac_code]  # ac_code is index of ac_label
            for k, ac_data in zip(['non_weighted', 'weighted'], [accessibility_data, accessibility_data_wt]):
                if ac_label not in self.Network.metrics['accessibility'][k]:
                    self.Network.metrics['accessibility'][k][ac_label] = {}
                for d_idx, d_key in enumerate(self.Network.distances):
                    self.Network.metrics['accessibility'][k][ac_label][d_key] = ac_data[ac_idx][d_idx]

        # unpack the numerical arrays
        if stats_keys:
            for num_idx, stats_key in enumerate(stats_keys):
                if stats_key not in self.Network.metrics['stats']:
                    self.Network.metrics['stats'][stats_key] = {}
                for k, stats_data in zip(['max',
                                          'min',
                                          'sum',
                                          'sum_weighted',
                                          'mean',
                                          'mean_weighted',
                                          'variance',
                                          'variance_weighted'],
                                         [stats_max,
                                          stats_min,
                                          stats_sum,
                                          stats_sum_wt,
                                          stats_mean,
                                          stats_mean_wt,
                                          stats_variance,
                                          stats_variance_wt]):
                    if k not in self.Network.metrics['stats'][stats_key]:
                        self.Network.metrics['stats'][stats_key][k] = {}
                    for d_idx, d_key in enumerate(self.Network.distances):
                        self.Network.metrics['stats'][stats_key][k][d_key] = stats_data[num_idx][d_idx]

    def hill_diversity(self,
                       landuse_labels: Union[list, tuple, np.ndarray],
                       qs: Union[list, tuple, np.ndarray] = None):
        """
        Compute hill diversity for the provided `landuse_labels` at the specified values of `q`. See
        [`DataLayer.compute_aggregated`](#datalayer-compute-aggregated) for additional information.

        Parameters
        ----------
        landuse_labels
            A set of land-use labels corresponding to the length and order of the data points. The labels should
            correspond to descriptors from the land-use schema, such as "retail" or "commercial".
        qs
            The values of `q` for which to compute Hill diversity, by default None

        Notes
        -----
        The data key is `hill`, e.g.:

        `NetworkLayer.metrics['mixed_uses']['hill'][<<q key>>][<<distance key>>][<<node idx>>]`
        """
        return self.compute_aggregated(landuse_labels, mixed_use_keys=['hill'], qs=qs)

    def hill_branch_wt_diversity(self,
                                 landuse_labels: Union[list, tuple, np.ndarray],
                                 qs: Union[list, tuple, np.ndarray] = None):
        """
        Compute distance-weighted hill diversity for the provided `landuse_labels` at the specified values of `q`. See
        [`DataLayer.compute_aggregated`](#datalayer-compute-aggregated) for additional information.

        Parameters
        ----------
        landuse_labels
            A set of land-use labels corresponding to the length and order of the data points. The labels should
            correspond to descriptors from the land-use schema, such as "retail" or "commercial".
        qs
            The values of `q` for which to compute Hill diversity, by default None

        Notes
        -----
        The data key is `hill_branch_wt`, e.g.:

        `NetworkLayer.metrics['mixed_uses']['hill_branch_wt'][<<q key>>][<<distance key>>][<<node idx>>]`
        """
        return self.compute_aggregated(landuse_labels, mixed_use_keys=['hill_branch_wt'], qs=qs)

    def compute_accessibilities(self,
                                landuse_labels: Union[list, tuple, np.ndarray],
                                accessibility_keys: Union[list, tuple]):
        """
        Compute land-use accessibilities for the specified land-use classification keys. See
        [`DataLayer.compute_aggregated`](#datalayer-compute-aggregated) for additional information.

        Parameters
        ----------
        landuse_labels
            A set of land-use labels corresponding to the length and order of the data points. The labels should
            correspond to descriptors from the land-use schema, such as "retail" or "commercial".
        accessibility_keys
            The land-use keys for which to compute accessibilies. The keys should be selected from the same land-use
            schema used for the `landuse_labels` parameter, e.g. "retail". The calculations will be performed in both
            `weighted` and `non_weighted` variants.

        Notes
        -----
        The data keys will correspond to the `accessibility_keys` specified, e.g. where computing `retail`
        accessibility:

        `NetworkLayer.metrics['accessibility']['weighted']['retail'][<<distance key>>][<<node idx>>]`<br>
        `NetworkLayer.metrics['accessibility']['non_weighted']['retail'][<<distance key>>][<<node idx>>]`
        """
        return self.compute_aggregated(landuse_labels, accessibility_keys=accessibility_keys)

    def compute_stats_single(self,
                             stats_key: str,
                             stats_data_arr: Union[list, tuple, np.ndarray]):
        """
        Compute stats for a single `stats_key` parameter. As with the landuse and mixed-use measures: stats will be
        computed for each distance that was specified when initialising the `NetworkLayer`.

        Parameters
        ----------
        stats_key
            A `str` key describing the stats computed for the `stats_data_arr` parameter. The computed stats will be
            saved to the `N.metrics` dictionary under this key.
        stats_data_arr
            A 1d `list`, `tuple` or `numpy` array of numerical data, where the length corresponds to the number of data
            points in the `DataLayer`.
        
        Notes
        -----
        The data key will correspond to the `stats_key` parameter, e.g. where using `occupants` as the key:
        
        `NetworkLayer.metrics['stats']['occupants'][<<stat type>>][<<distance key>>][<<node idx>>]`

        ::: tip Hint
        Per the above worked example, the following stat types will be available for each `stats_key` for each of the computed distances:
        - `max` and `min`
        - `sum` and `sum_weighted`
        - `mean` and `mean_weighted`
        - `variance` and `variance_weighted`
        :::
        """
        if stats_data_arr.ndim != 1:
            raise ValueError(
                'The stats_data_arr must be a single dimensional array with a length corresponding to the number of data points in the DataLayer.')
        return self.compute_aggregated(stats_keys=[stats_key], stats_data_arrs=[stats_data_arr])

    def compute_stats_multiple(self,
                               stats_keys: Union[list, tuple],
                               stats_data_arrs: Union[list, tuple, np.ndarray]):
        """
        

        Parameters
        ----------
        stats_keys
            A `list` or `tuple` of keys corresponding to the number of nested arrays passed to the `stats_data_arrs`
            parameter. The computed stats will be saved to the `N.metrics` dictionary under these keys.
        stats_data_arrs
            A 2d `list`, `tuple` or `numpy` array of numerical data, where the first dimension corresponds to the number
            of keys in the `stats_keys` parameter and the second dimension corresponds to number of data points in the
            `DataLayer`. See the below example.

        Notes
        -----
        ```python
        # for a DataLayer containg 5 data points
        stats_keys = ['valuations', 'floors', 'occupants']
        stats_data_arrs = [
            [50000, 60000, 55000, 42000, 46000],  # valuations
            [3, 3, 2, 3, 5],  # floors
            [420, 300, 220, 250, 600]  # occupants
        ]
        ```
        The data keys will correspond to the `stats_keys` parameter:

        `NetworkLayer.metrics['stats']['valuations'][<<stat type>>][<<distance key>>][<<node idx>>]`<br>
        `NetworkLayer.metrics['stats']['floors'][<<stat type>>][<<distance key>>][<<node idx>>]`<br>
        `NetworkLayer.metrics['stats']['occupants'][<<stat type>>][<<distance key>>][<<node idx>>]`
        """

        return self.compute_aggregated(stats_keys=stats_keys, stats_data_arrs=stats_data_arrs)

    def model_singly_constrained(self,
                                 key: str,
                                 i_data_map: np.ndarray,
                                 j_data_map: np.ndarray,
                                 i_weights: Union[list, tuple, np.ndarray],
                                 j_weights: Union[list, tuple, np.ndarray],
                                 angular: bool = False):
        """
        Undocumented method for computing singly-constrained interactions.
        """

        j_assigned, netw_flows = data.singly_constrained(self.Network._node_data,
                                                         self.Network._edge_data,
                                                         self.Network._node_edge_map,
                                                         distances=np.array(self.Network.distances),
                                                         betas=np.array(self.Network.betas),
                                                         i_data_map=i_data_map,
                                                         j_data_map=j_data_map,
                                                         i_weights=np.array(i_weights),
                                                         j_weights=np.array(j_weights),
                                                         angular=angular,
                                                         suppress_progress=checks.quiet_mode)

        # write the results to the Network's metrics dict
        if key not in self.Network.metrics['models']:
            self.Network.metrics['models'][key] = {}
        for d_idx, d_key in enumerate(self.Network.distances):
            if d_key not in self.Network.metrics['models'][key]:
                self.Network.metrics['models'][key][d_key] = {}
            self.Network.metrics['models'][key][d_key]['assigned_trips'] = j_assigned[d_idx]
            self.Network.metrics['models'][key][d_key]['network_flows'] = netw_flows[d_idx]


class DataLayerFromDict(DataLayer):
    """
    Directly transposes an appropriately prepared data dictionary into a `DataLayer`. This `class` calls
    [`data_map_from_dict`](#data-map-from-dict) internally. Methods and properties are inherited from the parent
    [`DataLayer`](#class-datalayer) class, which can be referenced for more information.
    """
    def __init__(self, data_dict: dict) -> DataLayer:
        """
        Parameters
        ----------
        data_dict
            A dictionary representing distinct data points, where each `key` represents a `uid` and each value
            represents a nested dictionary with `x` and `y` key-value pairs. The coordinates must be in a projected
            coordinate system matching that of the
            [`NetworkLayer`](/metrics/networks#class-networklayer) to which the data will
            be assigned.

        Returns
        -------
        DataLayer
            Returns a [`DataLayer`](#class-datalayer).

        Notes
        -----
        Example dictionary:
        ```python
        example_data_dict = {
            'uid_01': {
                'x': 6000956.463188213,
                'y': 600693.4059810264
            },
            'uid_02': {
                'x': 6000753.336609659,
                'y': 600758.7916663144
            }
        }
        ```
        """
        data_uids, data_map = data_map_from_dict(data_dict)

        super().__init__(data_uids, data_map)



'''
Centrality methods
'''
import logging
from typing import Union

import networkx as nx
import numpy as np
from numba.typed import Dict

from cityseer.algos import centrality, checks
from cityseer.tools import graphs

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def distance_from_beta(beta: Union[float, list, np.ndarray],
                       min_threshold_wt: float = checks.def_min_thresh_wt) -> np.ndarray:
    """
    Maps decay parameters $\beta$ to equivalent distance thresholds $d_{max}$ at the specified cutoff weight $w_{min}$.

    ::: warning Note
    It is generally not necessary to utilise this function directly. It will be called internally, if necessary, when invoking [`NetworkLayer`](#class-networklayer) or [`NetworkLayerFromNX`](#class-networklayerfromnx).
    :::

    Parameters
    ----------
    beta
        $\beta$ value/s to convert to distance thresholds $d_{max}$.
    min_threshold_wt
        An optional cutoff weight $w_{min}$ at which to set the distance threshold $d_{max}$, default of 0.01831563888873418.

    Returns
    -------
    np.ndarray
        A numpy array of distance thresholds $d_{max}$.

    Notes
    -----

    ```python
    from cityseer.metrics import networks
    # a list of betas
    betas = [-0.01, -0.02]
    # convert to distance thresholds
    d_max = networks.distance_from_beta(betas)
    print(d_max)
    # prints: array([400., 200.])
    ```

    Weighted measures such as the gravity index, weighted betweenness, and weighted land-use accessibilities are computed using a negative exponential decay function in the form of:

    $$weight = exp(\beta \cdot distance)$$

    The strength of the decay is controlled by the $\beta$ parameter, which reflects a decreasing willingness to walk correspondingly farther distances. For example, if $\beta=-0.005$ were to represent a person's willingness to walk to a bus stop, then a location $100m$ distant would be weighted at $60\%$ and a location $400m$ away would be weighted at $13.5\%$. After an initially rapid decrease, the weightings decay ever more gradually in perpetuity; thus, once a sufficiently small weight is encountered it becomes computationally expensive to consider locations any farther away. The minimum weight at which this cutoff occurs is represented by $w_{min}$, and the corresponding maximum distance threshold by $d_{max}$.

    ![Example beta decays](../.vitepress/plots/images/betas.png)
    
    [`NetworkLayer`](#class-networklayer) and [`NetworkLayerFromNX`](/metrics/networks#class-networklayerfromnx) can be invoked with either `distances` or `betas` parameters, but not both. If using the `betas` parameter, then this function will be called in order to extrapolate the distance thresholds implicitly, using:

    $$d_{max} = \frac{log\Big(w_{min}\Big)}{\beta}$$

    The default `min_threshold_wt` of $w_{min}=0.01831563888873418$ yields conveniently rounded $d_{max}$ walking thresholds, for example:

    | $\beta$ | $d_{max}$ |
    |-----------|---------|
    | $-0.02$ | $200m$ |
    | $-0.01$ | $400m$ |
    | $-0.005$ | $800m$ |
    | $-0.0025$ | $1600m$ |

    Overriding the default $w_{min}$ will adjust the $d_{max}$ accordingly, for example:

    | $\beta$ | $w_{min}$ | $d_{max}$ |
    |----------|----------|----------|
    | $-0.02$ | $0.01$ | $230m$ |
    | $-0.01$ | $0.01$ | $461m$ |
    | $-0.005$ | $0.01$ | $921m$ |
    | $-0.0025$ | $0.01$ | $1842m$ |

    """
    # cast to list form
    if isinstance(beta, (int, float)):
        beta = [beta]
    if not isinstance(beta, (list, tuple, np.ndarray)):
        raise TypeError('Please provide a beta or a list, tuple, or numpy.ndarray of betas.')
    # check that the betas do not have leading negatives
    for b in beta:
        if b > 0:
            raise ValueError('Please provide the beta value with the leading negative.')
        elif b == 0:
            # ints have no concept of -0, so catch betas that are positive 0 or -0 (in int form)
            # i.e. betas of -0.0 will successfully result in np.inf as opposed to -np.inf
            if np.log(min_threshold_wt) / b == -np.inf:
                raise ValueError('Please provide zeros in float form with a leading negative.')
    # cast to numpy
    beta = np.array(beta)
    # deduce the effective distance thresholds
    return np.log(min_threshold_wt) / beta


def beta_from_distance(distance: Union[float, list, np.ndarray],
                       min_threshold_wt: float = checks.def_min_thresh_wt) -> np.ndarray:
    """

    Maps distance thresholds $d_{max}$ to equivalent decay parameters $\beta$ at the specified cutoff weight $w_{min}$. See [`distance_from_beta`](#distance-from-beta) for additional discussion.

    ::: warning Note
    It is generally not necessary to utilise this function directly. It will be called internally, if necessary, when invoking [`NetworkLayer`](#class-networklayer) or [`NetworkLayerFromNX`](#class-networklayerfromnx).
    :::

    Parameters
    ----------
    distance
        $d_{max}$ value/s to convert to decay parameters $\beta$.
    min_threshold_wt
        The cutoff weight $w_{min}$ on which to model the decay parameters $\beta$, default of 0.01831563888873418.

    Returns
    -------
    np.ndarray
        A numpy array of decay parameters $\beta$.

    Notes
    -----
    
    ```python
    from cityseer.metrics import networks
    # a list of betas
    distances = [400, 200]
    # convert to betas
    betas = networks.beta_from_distance(distances)
    print(betas)  # prints: array([-0.01, -0.02])
    ```

    [`NetworkLayer`](#class-networklayer) and [`NetworkLayerFromNX`](#class-networklayerfromnx) can be invoked with either `distances` or `betas` parameters, but not both. If using the `distances` parameter, then this function will be called in order to extrapolate the decay parameters implicitly, using:

    $$\beta = \frac{log\Big(w_{min}\Big)}{d_{max}}$$

    The default `min_threshold_wt` of $w_{min}=0.01831563888873418$ yields conveniently rounded $\beta$ parameters, for example:

    | $d_{max}$ | $\beta$ |
    |-----------|---------|
    | $200m$ | $-0.02$ |
    | $400m$ | $-0.01$ |
    | $800m$ | $-0.005$ |
    | $1600m$ | $-0.0025$ |

    """
    # cast to list form
    if isinstance(distance, (int, float)):
        distance = [distance]
    if not isinstance(distance, (list, tuple, np.ndarray)):
        raise TypeError('Please provide a distance or a list, tuple, or numpy.ndarray of distances.')
    # check that the betas do not have leading negatives
    for d in distance:
        if d <= 0:
            raise ValueError('Please provide a positive distance value.')
    # cast to numpy
    distance = np.array(distance)
    # deduce the effective distance thresholds
    return np.log(min_threshold_wt) / distance


# TODO: sub-class network layer for nd; nd-ang; seg; seg-ang
# do all setup logic here in class and keep backend lean and modular

class NetworkLayer:
    """
    Network layers are used for network centrality computations and provide the backbone for landuse and statistical aggregations. [`NetworkLayerFromNX`](#class-networklayerfromnx) should be used instead if converting from a `NetworkX`
    `MultiGraph` to a `NetworkLayer`.

    A `NetworkLayer` requires either a set of distances $d_{max}$ or equivalent exponential decay parameters
    $\beta$, but not both. The unprovided parameter will be calculated implicitly in order to keep weighted and
    unweighted metrics in lockstep. The `min_threshold_wt` parameter can be used to generate custom mappings from
    one to the other: see [`distance_from_beta`](#distance-from-beta) for more information. These distances and betas
    are used for any subsequent centrality and land-use calculations.

    ```python
    from cityseer.metrics import networks
    from cityseer.tools import mock, graphs

    # prepare a mock graph
    G = mock.mock_graph()
    G = graphs.nX_simple_geoms(G)

    # if initialised with distances: 
    # betas for weighted metrics will be generated implicitly
    N = networks.NetworkLayerFromNX(G, distances=[200, 400, 800, 1600])
    print(N.distances)  # prints: [200, 400, 800, 1600]
    print(N.betas)  # prints: [-0.02, -0.01, -0.005, -0.0025]

    # if initialised with betas: 
    # distances for non-weighted metrics will be generated implicitly
    N = networks.NetworkLayerFromNX(G, betas=[-0.02, -0.01, -0.005, -0.0025])
    print(N.distances)  # prints: [200, 400, 800, 1600]
    print(N.betas)  # prints: [-0.02, -0.01, -0.005, -0.0025]
    ```

    There are two network centrality methods available depending on whether you're using a node-based or segment-based
    approach:

    - [`compute_node_centrality`](#networklayer-node-centrality)
    - [`compute_segment_centrality`](#networklayer-segment-centrality)

    These methods wrap the underlying `numba` optimised functions for computing centralities, and provides access to
    all of the underlying node-based or segment-based centrality methods. Multiple selected measures and distances are
    computed simultaneously to reduce the amount of time required for multi-variable and multi-scalar strategies.

    ::: tip Hints
    The reasons for picking one approach over another are varied:

    - Node based centralities compute the measures relative to each reachable node within the threshold distances. For
    this reason, they can be susceptible to distortions caused by messy graph topologies such redundant and varied
    concentrations of $degree=2$ nodes (e.g. to describe roadway geometry) or needlessly complex representations of
    street intersections. In these cases, the network should first be cleaned using methods such as those available in
    the [`graph`](/tools/graphs) module (see the [graph cleaning guide](/guide/cleaning) for examples). If a network
    topology has varied intensities of nodes but the street segments are less spurious, then segmentised methods can be
    preferable because they are based on segment distances: segment aggregations remain the same regardless of the
    number of intervening nodes, however, are not immune from situations such as needlessly complex representations of
    roadway intersections or a proliferation of walking paths in greenspaces;
    - Node-based `harmonic` centrality can be problematic on graphs where nodes are erroneously placed too close
    together or where impedances otherwise approach zero, as may be the case for simplest-path measures or small
    distance thesholds. This happens because the outcome of the division step can balloon towards $\infty$ once
    impedances decrease below $1$.
    - Note that `cityseer`'s implementation of simplest (angular) measures work on both primal (node or segment based)
    and dual graphs (node only).
    - Measures should only be directly compared on the same topology because different topologies can otherwise affect
    the expression of a measure. Accordingly, measures computed on dual graphs cannot be compared to measures computed
    on primal graphs because this does not account for the impact of differing topologies. Dual graph representations
    can have substantially greater numbers of nodes and edges for the same underlying street network; for example, a
    four-way intersection consisting of one node with four edges translates to four nodes and six edges on the dual.
    This effect is amplified for denser regions of the network.
    - Segmentised versions of centrality measures should not be computed on dual graph topologies because street segment
    lengths would be duplicated for each permutation of dual edge spanning street intersections. By way of example,
    the contribution of a single edge segment at a four-way intersection would be duplicated three times.
    - Global closeness is strongly discouraged because it does not behave suitably for localised graphs. Harmonic
    closeness or improved closeness should be used instead. Note that Global closeness ($\frac{nodes}{farness}$) and
    improved closeness ($\frac{nodes}{farness / nodes}$) can be recovered from the available metrics, if so desired,
    through additional (manual) steps.
    - Network decomposition can be a useful strategy when working at small distance thresholds, and confers advantages
    such as more regularly spaced snapshots and fewer artefacts at small distance thresholds where street edges
    intersect distance thresholds. However, the regular spacing of the decomposed segments will introduce spikes in the
    distributions of node-based centrality measures when working at very small distance thresholds. Segmentised versions
    may therefore be preferable when working at small thresholds on decomposed networks.
    :::

    The computed metrics will be written to a dictionary available at the `NetworkLayer.metrics` property and will be
    categorised by the respective centrality and distance keys:

    <small>`NetworkLayer.metrics['centrality'][<<measure key>>][<<distance key>>][<<node idx>>]`</small>

    For example, if `node_density`, and `node_betweenness_beta` centrality keys are computed at $800m$ and $1600m$,
    then the dictionary would assume the following structure:

    ```python
    # example structure
    NetworkLayer.metrics = {
        'centrality': {
            'node_density': {
                800: [<numpy array>],
                1600: [<numpy array>]
            },
            'node_betweenness_beta': {
                800: [<numpy array>],
                1600: [<numpy array>]
            }
        }
    }
    ```

    A worked example:

    ```python
    from cityseer.metrics import networks
    from cityseer.tools import mock, graphs

    # prepare a mock graph
    G = mock.mock_graph()
    G = graphs.nX_simple_geoms(G)

    # generate the network layer and compute some metrics
    N = networks.NetworkLayerFromNX(G, distances=[200, 400, 800, 1600])
    # compute a centrality measure
    N.node_centrality(measures=['node_density', 'node_betweenness_beta'])

    # fetch node density for 400m threshold for the first 5 nodes
    print(N.metrics['centrality']['node_density'][400][:5])
    # prints [15, 13, 10, 11, 12]

    # fetch betweenness beta for 1600m threshold for the first 5 nodes
    print(N.metrics['centrality']['node_betweenness_beta'][1600][:5])
    # prints [76.01161, 45.261307, 6.805982, 11.478158, 33.74703]
    ```

    The data can be handled using the underlying `numpy` arrays, and can also be unpacked to a dictionary using
    [`NetworkLayer.metrics_to_dict`](#networklayer-metrics-to-dict) or transposed to a `networkX` graph using
    [`NetworkLayer.to_networkX`](#networklayer-to-networkx).
    """

    def __init__(self,
                 node_uids: Union[list, tuple],
                 node_data: np.ndarray,
                 edge_data: np.ndarray,
                 node_edge_map: Dict,
                 distances: Union[list, tuple, np.ndarray] = None,
                 betas: Union[list, tuple, np.ndarray] = None,
                 min_threshold_wt: float = checks.def_min_thresh_wt):
        """
        Parameters
        ----------
        node_uids
            A `list` or `tuple` of node identifiers corresponding to each node. This list must be in the same order and
            of the same length as the `node_data`.
        node_data
            A 2d `numpy` array representing the graph's nodes. The indices of the second dimension correspond as
            follows:

            | idx | property |
            |-----|:---------|
            | 0 | `x` coordinate |
            | 1 | `y` coordinate |
            | 2 | `bool` describing whether the node is `live`. Metrics are only computed for `live` nodes. |

            The `x` and `y` node attributes determine the spatial coordinates of the node, and should be in a suitable
            projected (flat) coordinate reference system in metres. [`nX_wgs_to_utm`](/tools/graphs.html#nx-wgs-to-utm)
            can be used for converting a `networkX` graph from WGS84 `lng`, `lat` geographic coordinates to the local
            UTM `x`, `y` projected coordinate system.

            When calculating local network centralities or land-use accessibilities, it is best-practice to buffer the
            network by a distance equal to the maximum distance threshold to be considered. This prevents problematic
            results arising due to boundary roll-off effects.
            
            The `live` node attribute identifies nodes falling within the areal boundary of interest as opposed to those
            that fall within the surrounding buffered area. Calculations are only performed for `live=True` nodes, thus
            reducing frivolous computation while also cleanly identifying which nodes are in the buffered roll-off area.
            If some other process will be used for filtering the nodes, or if boundary roll-off is not being considered,
            then set all nodes to `live=True`.
        edge_data
            A 2d `numpy` array representing the graph's edges. Each edge will be described separately for each direction
            of travel. The indices of the second dimension correspond as follows:
            | idx | property |
            |-----|:---------|
            | 0 | start node `idx` |
            | 1 | end node `idx` |
            | 2 | the segment length in metres |
            | 3 | the sum of segment's angular change |
            | 4 | an 'impedance factor' which can be applied to magnify or reduce the effect of the edge's impedance on
            shortest-path calculations. e.g. for gradients or other such considerations. Use with caution. |
            | 5 | the edge's entry angular bearing |
            | 6 | the edge's exit angular bearing |

            The start and end edge `idx` attributes point to the corresponding node indices in the `node_data` array.

            The `length` edge attribute (index $2$) should always correspond to the edge lengths in metres. This is used
            when calculating the distances traversed by the shortest-path algorithm so that the respective $d_{max}$
            maximum distance thresholds can be enforced: these distance thresholds are based on the actual network-paths
            traversed by the algorithm as opposed to crow-flies distances.

            The `angle_sum` edge bearing (index $3$) should correspond to the total angular change along the length of
            the segment. This is used when calculating angular impedances for simplest-path measures. The
            `start_bearing` (index $5$) and `end_bearing` (index $6$) attributes respectively represent the starting and
            ending bearing of the segment. This is also used when calculating simplest-path measures when the algorithm
            steps from one edge to another.

            The `imp_factor` edge attribute (index $4$) represents an impedance multiplier for increasing or diminishing
            the impedance of an edge. This is ordinarily set to $1$, therefor not impacting calculations. By setting
            this to greater or less than $1$, the edge will have a correspondingly higher or lower impedance. This can
            be used to take considerations such as street gradients into account, but should be used with caution.
        node_edge_map
            A `numba` `Dict` with `node_data` indices as keys and `numba` `List` types as values containing the out-edge
            indices for each node.
        distances
            A distance, or `list`, `tuple`, or `numpy` array of distances corresponding to the local $d_{max}$
            thresholds to be used for centrality (and land-use) calculations. The $\beta$ parameters (for
            distance-weighted metrics) will be determined implicitly. If the `distances` parameter is not provided, then
            the `beta` parameter must be provided instead. Use a distance of `np.inf` where no distance threshold should
            be enforced.
        betas
            A $\beta$, or `list`, `tuple`, or `numpy` array of $\beta$ to be used for the exponential decay function for
            weighted metrics. The `distance` parameters for unweighted metrics will be determined implicitly. If the
            `betas` parameter is not provided, then the `distance` parameter must be provided instead.
        min_threshold_wt
            The default `min_threshold_wt` parameter can be overridden to generate custom mappings between the
            `distance` and `beta` parameters. See [`distance_from_beta`](#distance-from-beta) for more information.

        Returns
        -------
        NetworkLayer
            A `NetworkLayer`.

        Notes
        -----
        ::: tip Hint
        It is possible to represent unlimited $d_{max}$ distance thresholds by setting one of the specified `distance`
        parameter values to `np.inf`. Note that this may substantially increase the computational time required for the
        completion of the algorithms on large networks.
        :::

        Properties
        ----------
        """
        self._uids = node_uids
        self._node_data = node_data
        self._edge_data = edge_data
        self._node_edge_map = node_edge_map
        self._distances = distances
        self._betas = betas
        self._min_threshold_wt = min_threshold_wt
        self.metrics = {
            'centrality': {},
            'mixed_uses': {},
            'accessibility': {
                'non_weighted': {},
                'weighted': {}
            },
            'stats': {},
            'models': {}
        }
        # for storing originating networkX graph
        self._networkX_multigraph = None
        # check the data structures
        if len(self._uids) != len(self._node_data):
            raise ValueError('The number of indices does not match the number of nodes.')
        # check network maps
        checks.check_network_maps(self._node_data, self._edge_data, self._node_edge_map)
        # if distances, check the types and generate the betas
        if self._distances is not None and self._betas is None:
            if isinstance(self._distances, (int, float)):
                self._distances = [self._distances]
            if isinstance(self._distances, (list, tuple, np.ndarray)):
                if len(self._distances) == 0:
                    raise ValueError('Please provide at least one distance.')
            else:
                raise TypeError('Please provide a distance, or a list, tuple, or numpy.ndarray of distances.')
            # generate the betas
            self._betas = beta_from_distance(self._distances,
                                             min_threshold_wt=self._min_threshold_wt)
        # if betas, generate the distances
        elif self._betas is not None and self._distances is None:
            if isinstance(self._betas, (float)):
                self._betas = [self._betas]
            if isinstance(self._betas, (list, tuple, np.ndarray)):
                if len(self._betas) == 0:
                    raise ValueError('Please provide at least one beta.')
            else:
                raise TypeError('Please provide a beta, or a list, tuple, or numpy.ndarray of betas.')
            self._distances = distance_from_beta(self._betas,
                                                 min_threshold_wt=self._min_threshold_wt)
        else:
            raise ValueError('Please provide either distances or betas, but not both.')

    @property
    def uids(self):
        """Unique ids corresponding to each node in the graph's node_map."""
        return self._uids

    @property
    def distances(self):
        """The distance threshold/s at which the class has been initialised."""
        return self._distances

    @property
    def betas(self):
        """The distance decay $\beta$ thresholds (spatial impedance) at which the class is initialised."""
        return self._betas

    @property
    def node_x_arr(self):
        return self._node_data[:, 0]

    @property
    def node_y_arr(self):
        return self._node_data[:, 1]

    @property
    def node_live_arr(self):
        return self._node_data[:, 2]

    @property
    def edge_lengths_arr(self):
        return self._edge_data[:, 2]

    @property
    def edge_angles_arr(self):
        return self._edge_data[:, 3]

    @property
    def edge_impedance_factors_arr(self):
        return self._edge_data[:, 4]

    @property
    def edge_in_bearings_arr(self):
        return self._edge_data[:, 5]

    @property
    def edge_out_bearings_arr(self):
        return self._edge_data[:, 6]

    @property
    def networkX_multigraph(self):
        """If initialised with `NetworkLayerFromNX`, the `networkX` `MultiGraph` from which the graph is derived."""
        return self._networkX_multigraph

    @networkX_multigraph.setter
    def networkX_multigraph(self, networkX_multigraph):
        self._networkX_multigraph = networkX_multigraph

    # for retrieving metrics to a dictionary
    def metrics_to_dict(self):
        """
        Unpacks all calculated metrics from the `NetworkLayer.metrics` property into a `python` dictionary. The
        dictionary `keys` will correspond to the node `uids`.

        Notes
        -----
        ```python
        from cityseer.metrics import networks
        from cityseer.tools import mock, graphs

        # prepare a mock graph
        G = mock.mock_graph()
        G = graphs.nX_simple_geoms(G)

        # generate the network layer and compute some metrics
        N = networks.NetworkLayerFromNX(G, distances=[200, 400, 800, 1600])
        N.node_centrality(measures=['node_harmonic'])

        # let's select a random node id
        random_idx = 6
        random_uid = N.uids[random_idx]

        # the data is directly available at N.metrics
        # in this case the data is stored in arrays corresponding to the node indices
        print(N.metrics['centrality']['node_harmonic'][200][random_idx])
        # prints: 0.023120252

        # let's convert the data to a dictionary
        # the unpacked data is now stored by the uid of the node identifier
        data_dict = N.metrics_to_dict()
        print(data_dict[random_uid]['centrality']['node_harmonic'][200])
        # prints: 0.023120252
        ```
        """
        m = {}
        for i, uid in enumerate(self._uids):
            m[uid] = {
                'x': self.node_x_arr[i],
                'y': self.node_y_arr[i],
                'live': self.node_live_arr[i] == 1
            }
            # unpack centralities
            m[uid]['centrality'] = {}
            for m_key, m_val in self.metrics['centrality'].items():
                m[uid]['centrality'][m_key] = {}
                for d_key, d_val in m_val.items():
                    m[uid]['centrality'][m_key][d_key] = d_val[i]

            m[uid]['mixed_uses'] = {}
            for m_key, m_val in self.metrics['mixed_uses'].items():
                m[uid]['mixed_uses'][m_key] = {}
                if 'hill' in m_key:
                    for q_key, q_val in m_val.items():
                        m[uid]['mixed_uses'][m_key][q_key] = {}
                        for d_key, d_val in q_val.items():
                            m[uid]['mixed_uses'][m_key][q_key][d_key] = d_val[i]
                else:
                    for d_key, d_val in m_val.items():
                        m[uid]['mixed_uses'][m_key][d_key] = d_val[i]
            m[uid]['accessibility'] = {
                'non_weighted': {},
                'weighted': {}
            }
            for cat in ['non_weighted', 'weighted']:
                for cl_key, cl_val in self.metrics['accessibility'][cat].items():
                    m[uid]['accessibility'][cat][cl_key] = {}
                    for d_key, d_val in cl_val.items():
                        m[uid]['accessibility'][cat][cl_key][d_key] = d_val[i]
            m[uid]['stats'] = {}
            for th_key, th_val in self.metrics['stats'].items():
                m[uid]['stats'][th_key] = {}
                for stat_key, stat_val in th_val.items():
                    m[uid]['stats'][th_key][stat_key] = {}
                    for d_key, d_val in stat_val.items():
                        m[uid]['stats'][th_key][stat_key][d_key] = d_val[i]
        return m

    # for unpacking to a networkX graph
    def to_networkX(self) -> nx.MultiGraph:
        """
        Transposes a `NetworkLayer` into a `networkX` `MultiGraph`. This method calls
        [`nX_from_graph_maps`](/tools/graphs.html#nx-from-graph-maps) internally.

        Returns
        -------
        nx.MultiGraph
            A `networkX` `MultiGraph`.
            
            `x`, `y`, and `live` node attributes will be copied from `node_data` to the `MultiGraph` nodes. `length`,
            `angle_sum`, `imp_factor`, `start_bearing`, and `end_bearing` attributes will be copied from the `edge_data`
            to the `MultiGraph` edges.
            
            If a `metrics_dict` is provided, all derived data will be copied to the `MultiGraph` nodes based on matching
            node identifiers.

        Notes
        -----
        ```python
        from cityseer.metrics import networks
        from cityseer.tools import mock, graphs

        # prepare a mock graph
        G = mock.mock_graph()
        G = graphs.nX_simple_geoms(G)

        # generate the network layer and compute some metrics
        N = networks.NetworkLayerFromNX(G, distances=[200, 400, 800, 1600])
        # compute some-or-other metrics
        N.node_centrality(measures=['node_harmonic'])
        # convert back to networkX
        G_post = N.to_networkX()

        # let's select a random node id
        random_idx = 6
        random_uid = N.uids[random_idx]

        print(N.metrics['centrality']['node_harmonic'][200][random_idx])
        # prints: 0.023120252

        # the metrics have been copied to the new networkX graph
        print(G_post.nodes[random_uid]['metrics']['centrality']['node_harmonic'][200])
        # prints: 0.023120252
        ```
        ![Graph before conversion](../.vitepress/plots/images/graph_before.png)
        _A `networkX` graph before conversion to a `NetworkLayer`._
        ![Graph after conversion back to networkX](../.vitepress/plots/images/graph_after.png)
        _A graph after conversion back to `networkX`._
        """
        metrics_dict = self.metrics_to_dict()
        return graphs.nX_from_graph_maps(self._uids,
                                         self._node_data,
                                         self._edge_data,
                                         self._node_edge_map,
                                         self._networkX_multigraph,
                                         metrics_dict)

    # deprecated method
    def compute_centrality(self, **kwargs):
        """
        This method is deprecated and, if invoked, will raise a DeprecationWarning. Please use
        [`compute_node_centrality`](#networklayer-node-centrality) or [`compute_segment_centrality`](#networklayer-segment-centrality)
        instead.

        Raises
        ------
        DeprecationWarning
            
        """
        raise DeprecationWarning('The compute_centrality method has been deprecated. '
                                 'It has been split into two: '
                                 'use "node_centrality" for node based measures '
                                 'and "segment_centrality" for segmentised measures.'
                                 'See the documentation for further information.')

    # provides access to the underlying centrality.local_centrality method
    def node_centrality(self,
                        measures: Union[list, tuple] = None,
                        angular: bool = False):
        """


        Parameters
        ----------
        measures
            A list or tuple of strings, containing any combination of the following `key` values, computed within the
            respective distance thresholds of $d_{max}$.
        angular
            A boolean indicating whether to use shortest or simplest path heuristics, by default False

        Notes
        -----

        The following keys use the shortest-path heuristic, and are available when the `angular` parameter is set to the default value of `False`:

        | key                   | formula | notes |
        | ----------------------| :------:| ----- |
        | node_density          | $\scriptstyle\sum_{j\neq{i}}^{n}1$ | A summation of nodes. |
        | node_farness          | $\scriptstyle\sum_{j\neq{i}}^{n}d_{(i,j)}$ | A summation of distances in metres. |
        | node_cycles           | $\scriptstyle\sum_{j\neq{i}j=cycle}^{n}1$ | A summation of network cycles. |
        | node_harmonic         | $\scriptstyle\sum_{j\neq{i}}^{n}\frac{1}{Z_{(i,j)}}$ | Harmonic closeness is an
        appropriate form of closeness centrality for localised implementations constrained by the threshold $d_{max}$. |
        | node_beta             | $\scriptstyle\sum_{j\neq{i}}^{n}\\exp(\beta\cdot d[i,j])$ | Also known as the
        '_gravity index_'. This is a spatial impedance metric differentiated from other closeness centralities by the
        use of an explicit $\beta$ parameter, which can be used to model the decay in walking tolerance as distances
        increase. |
        | node_betweenness      | $\scriptstyle\sum_{j\neq{i}}^{n}\sum_{k\neq{j}\neq{i}}^{n}1$ | Betweenness centrality
        summing all shortest-paths traversing each node $i$. |
        | node_betweenness_beta | $\scriptstyle\sum_{j\neq{i}}^{n}\sum_{k\neq{j}\neq{i}}^{n}\\exp(\beta\cdot d[j,k])$ | Applies
        a spatial impedance decay function to betweenness centrality. $d$ represents the full distance from any $j$ to
        $k$ node pair passing through node $i$. |

        The following keys use the simplest-path (shortest-angular-path) heuristic, and are available when the `angular`
        parameter is explicitly set to `True`:

        | key                      | formula | notes |
        | ------------------------ | :-----: | ----- |
        | node_harmonic_angular    | $\scriptstyle\sum_{j\neq{i}}^{n}\frac{1}{Z_{(i,j)}}$ | The simplest-path
        implementation of harmonic closeness uses angular-distances for the impedance parameter. Angular-distances are
        normalised by 180 and added to $1$ to avoid division by zero: $\scriptstyle{Z = 1 + (angularchange/180)}$. |
        | node_betweenness_angular | $\scriptstyle\sum_{j\neq{i}}^{n}\sum_{k\neq{j}\neq{i}}^{n}1$ | The simplest-path
        version of betweenness centrality. This is distinguished from the shortest-path version by use of a
        simplest-path heuristic (shortest angular distance). |
        """
        # see centrality.local_centrality for integrity checks on closeness and betweenness keys
        # typos are caught below
        if not angular:
            heuristic = 'shortest (non-angular)'
            options = (
                'node_density',
                'node_farness',
                'node_cycles',
                'node_harmonic',
                'node_beta',
                'node_betweenness',
                'node_betweenness_beta'
            )
        else:
            heuristic = 'simplest (angular)'
            options = (
                'node_harmonic_angular',
                'node_betweenness_angular'
            )
        if measures is None:
            raise ValueError(f'Please select at least one measure to compute.')
        measure_keys = []
        for measure in measures:
            if measure not in options:
                raise ValueError(f'Invalid network measure: {measure}. '
                                 f'Must be one of {", ".join(options)} when using {heuristic} path heuristic.')
            if measure in measure_keys:
                raise ValueError(f'Please remove duplicate measure: {measure}.')
            measure_keys.append(measure)
        measure_keys = tuple(measure_keys)
        if not checks.quiet_mode:
            logger.info(f'Computing {", ".join(measure_keys)} centrality measures using {heuristic} path heuristic.')
        measures_data = centrality.local_node_centrality(
            self._node_data,
            self._edge_data,
            self._node_edge_map,
            np.array(self._distances),
            np.array(self._betas),
            measure_keys,
            angular,
            suppress_progress=checks.quiet_mode)
        # write the results
        # writing metrics to dictionary will check for pre-existing
        # but writing sub-distances arrays will overwrite prior
        for measure_idx, measure_name in enumerate(measure_keys):
            if measure_name not in self.metrics['centrality']:
                self.metrics['centrality'][measure_name] = {}
            for d_idx, d_key in enumerate(self._distances):
                self.metrics['centrality'][measure_name][d_key] = measures_data[measure_idx][d_idx]

    # provides access to the underlying centrality.local_centrality method
    def segment_centrality(self,
                           measures: Union[list, tuple] = None,
                           angular: bool = False):
        """
        A list or tuple of strings, containing any combination of the following `key` values, computed within the
        respective distance thresholds of $d_{max}$.

        Parameters
        ----------
        measures
            A list or tuple of strings, containing any combination of the following `key` values, computed within the
            respective distance thresholds of $d_{max}$.
        angular
            A boolean indicating whether to use shortest or simplest path heuristics, by default False

        Notes
        -----
        The following keys use the shortest-path heuristic, and are available when the `angular` parameter is set to the
        default value of `False`:

        | key                 | formula | notes |
        | ------------------- | :-----: |------ |
        | segment_density     | $\scriptstyle\sum_{(a, b)}^{edges}d_{b} - d_{a}$ | A summation of edge lengths. |
        | segment_harmonic    | $\scriptstyle\sum_{(a, b)}^{edges}\int_{a}^{b}\ln(b) -\ln(a)$ | A continuous form of
        harmonic closeness centrality applied to edge lengths. |
        | segment_beta        | $\scriptstyle\sum_{(a, b)}^{edges}\int_{a}^{b}\frac{\exp(\beta\cdot b) -\exp(\beta\cdot a)}{\beta}$ | A
        continuous form of beta-weighted (gravity index) centrality applied to edge lengths. |
        | segment_betweenness | | A continuous form of betweenness: Resembles `segment_beta` applied to edges situated
        on shortest paths between all nodes $j$ and $k$ passing through $i$. |

        The following keys use the simplest-path (shortest-angular-path) heuristic, and are available when the `angular`
        parameter is explicitly set to `True`.

        | key                       | formula | notes |
        | ------------------------- | :-----: | ----- |
        | segment_harmonic_hybrid   | $\scriptstyle\sum_{(a, b)}^{edges}\frac{d_{b} - d_{a}}{Z}$ | Weights angular
        harmonic centrality by the lengths of the edges. See `node_harmonic_angular`. |
        | segment_betweeness_hybrid | | A continuous form of angular betweenness: Resembles `segment_harmonic_hybrid`
        applied to edges situated on shortest paths between all nodes $j$ and $k$ passing through $i$. |

        """
        # see centrality.local_centrality for integrity checks on closeness and betweenness keys
        # typos are caught below
        if not angular:
            heuristic = 'shortest (non-angular)'
            options = (
                'segment_density',
                'segment_harmonic',
                'segment_beta',
                'segment_betweenness'
            )
        else:
            heuristic = 'simplest (angular)'
            options = (
                'segment_harmonic_hybrid',
                'segment_betweeness_hybrid'
            )
        if measures is None:
            raise ValueError(f'Please select at least one measure to compute.')
        measure_keys = []
        for measure in measures:
            if measure not in options:
                raise ValueError(f'Invalid network measure: {measure}. '
                                 f'Must be one of {", ".join(options)} when using {heuristic} path heuristic.')
            if measure in measure_keys:
                raise ValueError(f'Please remove duplicate measure: {measure}.')
            measure_keys.append(measure)
        measure_keys = tuple(measure_keys)
        if not checks.quiet_mode:
            logger.info(f'Computing {", ".join(measure_keys)} centrality measures using {heuristic} path heuristic.')
        measures_data = centrality.local_segment_centrality(
            self._node_data,
            self._edge_data,
            self._node_edge_map,
            np.array(self._distances),
            np.array(self._betas),
            measure_keys,
            angular,
            suppress_progress=checks.quiet_mode)
        # write the results
        # writing metrics to dictionary will check for pre-existing
        # but writing sub-distances arrays will overwrite prior
        for measure_idx, measure_name in enumerate(measure_keys):
            if measure_name not in self.metrics['centrality']:
                self.metrics['centrality'][measure_name] = {}
            for d_idx, d_key in enumerate(self._distances):
                self.metrics['centrality'][measure_name][d_key] = measures_data[measure_idx][d_idx]


class NetworkLayerFromNX(NetworkLayer):
    def __init__(self,
                 networkX_multigraph: nx.MultiGraph,
                 distances: Union[list, tuple, np.ndarray] = None,
                 betas: Union[list, tuple, np.ndarray] = None,
                 min_threshold_wt: float = checks.def_min_thresh_wt) -> NetworkLayer:
        """
        Directly transposes a `networkX` `MultiGraph` into a `NetworkLayer`. This `class` simplifies the conversion of
        a `NetworkX` `MultiGraph` by calling [`graph_maps_from_nX`](/tools/graphs.html#graph-maps-from-nx) internally.
        Methods and properties are inherited from the parent [`NetworkLayer`](#class-networklayer) class.

        Parameters
        ----------
        networkX_multigraph
            A `networkX` `MultiGraph`.
            
            `x` and `y` node attributes are required. The `live` node attribute is optional, but recommended. See
            [`NetworkLayer`](#class-networklayer) for more information about what these attributes represent.
        distances
            See [`NetworkLayer`](#networklayer).
        betas
            See [`NetworkLayer`](#networklayer).
        min_threshold_wt
            See [`NetworkLayer`](#networklayer).

        Returns
        -------
        NetworkLayer
            A `NetworkLayer`.
        """
        node_uids, node_data, edge_data, node_edge_map = graphs.graph_maps_from_nX(networkX_multigraph)
        super().__init__(node_uids,
                         node_data,
                         edge_data,
                         node_edge_map,
                         distances,
                         betas,
                         min_threshold_wt)
        # keep reference to networkX graph
        self.networkX_multigraph = networkX_multigraph



from typing import Tuple

import numpy as np
from numba import njit
from numba.typed import Dict

from cityseer.algos import centrality, checks, diversity


@njit(cache=True)
def radial_filter(src_x: float, src_y: float, x_arr: np.ndarray, y_arr: np.ndarray, max_dist: float) -> np.ndarray:
    if len(x_arr) != len(y_arr):
        raise ValueError('Mismatching x and y array lengths.')

    # filter by distance
    total_count = len(x_arr)
    data_filter = np.full(total_count, False)

    # if infinite max, then no need to check distances
    if max_dist == np.inf:
        data_filter[:] = True
        return data_filter

    else:
        for i in range(total_count):
            dist = np.hypot(x_arr[i] - src_x, y_arr[i] - src_y)
            if dist <= max_dist:
                data_filter[i] = True

    return data_filter


@njit(cache=True)
def find_nearest(src_x: float, src_y: float, x_arr: np.ndarray, y_arr: np.ndarray, max_dist: float) -> Tuple[
    int, float]:
    if len(x_arr) != len(y_arr):
        raise ValueError('Mismatching x and y array lengths.')

    # filter by distance
    total_count = len(x_arr)
    min_idx = np.nan
    min_dist = np.inf

    for i in range(total_count):
        dist = np.hypot(x_arr[i] - src_x, y_arr[i] - src_y)
        if dist <= max_dist and dist < min_dist:
            min_idx = i
            min_dist = dist

    return min_idx, min_dist


@njit(cache=False)
def assign_to_network(data_map: np.ndarray,
                      node_data: np.ndarray,
                      edge_data: np.ndarray,
                      node_edge_map: Dict,
                      max_dist: float,
                      suppress_progress: bool = False) -> np.ndarray:
    '''
    To save unnecessary computation - this is done once and written to the data map.

    1 - find the closest network node from each data point
    2A - wind clockwise along the network to preferably find a block cycle surrounding the node
    2B - in event of topological traps, try anti-clockwise as well
    3A - select the closest block cycle node
    3B - if no enclosing cycle - simply use the closest node
    4 - find the neighbouring node that minimises the distance between the data point on "street-front"
    NODE MAP:
    0 - x
    1 - y
    2 - live
    EDGE MAP:
    0 - start node
    1 - end node
    2 - length in metres
    3 - sum of angular travel along length
    4 - impedance factor
    5 - entry bearing
    6 - exit bearing
    DATA MAP:
    0 - x
    1 - y
    2 - assigned network index - nearest
    3 - assigned network index - next-nearest
    '''
    checks.check_network_maps(node_data, edge_data, node_edge_map)

    def calculate_rotation(point_a, point_b):
        # https://stackoverflow.com/questions/37459121/calculating-angle-between-three-points-but-only-anticlockwise-in-python
        # these two points / angles are relative to the origin - so pass in difference between the points and origin as vectors
        ang_a = np.arctan2(point_a[1], point_a[0])  # arctan is in y/x order
        ang_b = np.arctan2(point_b[1], point_b[0])
        return np.rad2deg((ang_a - ang_b) % (2 * np.pi))

    def calculate_rotation_smallest(point_a, point_b):
        # smallest difference angle
        ang_a = np.rad2deg(np.arctan2(point_a[1], point_a[0]))
        ang_b = np.rad2deg(np.arctan2(point_b[1], point_b[0]))
        return np.abs((ang_b - ang_a + 180) % 360 - 180)

    def road_distance(d_coords, netw_idx_a, netw_idx_b):
        a_coords = node_data[netw_idx_a, :2]
        b_coords = node_data[netw_idx_b, :2]
        # get the angles from either intersection node to the data point
        ang_a = calculate_rotation_smallest(d_coords - a_coords, b_coords - a_coords)
        ang_b = calculate_rotation_smallest(d_coords - b_coords, a_coords - b_coords)
        # assume offset street segment if either is significantly greater than 90 (in which case sideways offset from road)
        if ang_a > 110 or ang_b > 110:
            return np.inf, np.nan, np.nan
        # calculate height from two sides and included angle
        side_a = np.hypot(d_coords[0] - a_coords[0], d_coords[1] - a_coords[1])
        side_b = np.hypot(d_coords[0] - b_coords[0], d_coords[1] - b_coords[1])
        base = np.hypot(a_coords[0] - b_coords[0], a_coords[1] - b_coords[1])
        # forestall potential division by zero
        if base == 0:
            return np.inf, np.nan, np.nan
        # heron's formula
        s = (side_a + side_b + base) / 2  # perimeter / 2
        a = np.sqrt(s * (s - side_a) * (s - side_b) * (s - base))
        # area is 1/2 base * h, so h = area / (0.5 * base)
        h = a / (0.5 * base)
        # NOTE - the height of the triangle may be less than the distance to the nodes
        # happens due to offset segments: can cause wrong assignment where adjacent segments have same triangle height
        # in this case, set to length of closest node so that h (minimum distance) is still meaningful
        # return indices in order of nearest then next nearest
        if side_a < side_b:
            if ang_a > 90:
                h = side_a
            return h, netw_idx_a, netw_idx_b
        else:
            if ang_b > 90:
                h = side_b
            return h, netw_idx_b, netw_idx_a

    def closest_intersections(d_coords, pr_map, end_node):
        if len(pr_map) == 1:
            return np.inf, end_node, np.nan
        current_idx = end_node
        next_idx = int(pr_map[int(end_node)])
        if len(pr_map) == 2:
            return road_distance(d_coords, current_idx, next_idx)
        nearest_idx = np.nan
        next_nearest_idx = np.nan
        min_d = np.inf
        first_pred = next_idx  # for finding end of loop
        while True:
            h, n_idx, n_n_idx = road_distance(d_coords, current_idx, next_idx)
            if h < min_d:
                min_d = h
                nearest_idx = n_idx
                next_nearest_idx = n_n_idx
            # if the next in the chain is nan, then break
            if np.isnan(pr_map[next_idx]):
                break
            current_idx = next_idx
            next_idx = int(pr_map[next_idx])
            if next_idx == first_pred:
                break
        return min_d, nearest_idx, next_nearest_idx

    pred_map = np.full(len(node_data), np.nan)
    netw_coords = node_data[:, :2]
    netw_x_arr = node_data[:, 0]
    netw_y_arr = node_data[:, 1]
    data_coords = data_map[:, :2]
    data_x_arr = data_map[:, 0]
    data_y_arr = data_map[:, 1]
    total_count = len(data_map)
    # setup progress bar params
    steps = int(total_count / 10000)
    for data_idx in range(total_count):
        if not suppress_progress:
            checks.progress_bar(data_idx, total_count, steps)
        # find the nearest network node
        min_idx, min_dist = find_nearest(data_x_arr[data_idx], data_y_arr[data_idx], netw_x_arr, netw_y_arr, max_dist)
        # in some cases no network node will be within max_dist... so accept NaN
        if np.isnan(min_idx):
            continue
        # nearest is initially set for this nearest node, but if a nearer street-edge is found, it will be overriden
        nearest = min_idx
        next_nearest = np.nan
        # set start node to nearest network node
        node_idx = int(min_idx)
        # keep track of visited nodes
        pred_map.fill(np.nan)
        # state
        reversing = False
        # keep track of previous indices
        prev_idx = np.nan
        # iterate neighbours
        while True:
            # reset neighbour rotation and index counters
            rotation = np.nan
            nb_idx = np.nan
            # iterate the edges
            for edge_idx in node_edge_map[node_idx]:
                # get the edge's start and end node indices
                start, end = edge_data[edge_idx, :2]
                # cast to int for indexing
                new_idx = int(end)
                # don't follow self-loops
                if new_idx == node_idx:
                    continue
                # check that this isn't the previous node (already visited as neighbour from other direction)
                if np.isfinite(prev_idx) and new_idx == prev_idx:
                    continue
                # look for the new neighbour with the smallest rightwards (anti-clockwise arctan2) angle
                # measure the angle relative to the data point for the first node
                if np.isnan(prev_idx):
                    r = calculate_rotation(netw_coords[int(new_idx)] - netw_coords[node_idx],
                                           data_coords[data_idx] - netw_coords[node_idx])
                # else relative to the previous node
                else:
                    r = calculate_rotation(netw_coords[int(new_idx)] - netw_coords[node_idx],
                                           netw_coords[int(prev_idx)] - netw_coords[node_idx])
                if reversing:
                    r = 360 - r
                # if least angle, update
                if np.isnan(rotation) or r < rotation:
                    rotation = r
                    nb_idx = new_idx
            # allow backtracking if no neighbour is found - i.e. dead-ends
            if np.isnan(nb_idx):
                if np.isnan(pred_map[node_idx]):
                    # for isolated nodes: nb_idx == np.nan, pred_map[node_idx] == np.nan, and prev_idx == np.nan
                    if np.isnan(prev_idx):
                        break
                    # for isolated edges, the algorithm gets turned-around back to the starting node with nowhere to go
                    # nb_idx == np.nan, pred_map[node_idx] == np.nan
                    # in these cases, pass closest_intersections the prev idx so that it has a predecessor to follow
                    d, n, n_n = closest_intersections(data_coords[data_idx], pred_map, int(prev_idx))
                    if d < min_dist:
                        nearest = n
                        next_nearest = n_n
                    break
                # otherwise, go ahead and backtrack
                nb_idx = pred_map[node_idx]
            # if the distance is exceeded, reset and attempt in the other direction
            dist = np.hypot(netw_x_arr[int(nb_idx)] - data_x_arr[data_idx],
                            netw_y_arr[int(nb_idx)] - data_y_arr[data_idx])
            if dist > max_dist:
                pred_map[int(nb_idx)] = node_idx
                d, n, n_n = closest_intersections(data_coords[data_idx], pred_map, int(nb_idx))
                # if the distance to the street edge is less than the nearest node, or than the prior closest edge
                if d < min_dist:
                    min_dist = d
                    nearest = n
                    next_nearest = n_n
                # reverse and try in opposite direction
                if not reversing:
                    reversing = True
                    pred_map.fill(np.nan)
                    node_idx = int(min_idx)
                    prev_idx = np.nan
                    continue
                break
            # ignore the following conditions while backtracking
            # (if backtracking, the current node's predecessor will be equal to the new neighbour)
            if nb_idx != pred_map[node_idx]:
                # if the new nb node has already been visited then terminate, this prevent infinite loops
                # or, if the algorithm has circled the block back to the original starting node
                if not np.isnan(pred_map[int(nb_idx)]) or nb_idx == min_idx:
                    # set the final predecessor, BUT ONLY if re-encountered the original node
                    # this would otherwise occlude routes (e.g. backtracks) that have passed the same node twice
                    # (such routes are still able to recover the closest edge)
                    if nb_idx == min_idx:
                        pred_map[int(nb_idx)] = node_idx
                    d, n, n_n = closest_intersections(data_coords[data_idx], pred_map, int(nb_idx))
                    if d < min_dist:
                        nearest = n
                        next_nearest = n_n
                    break
                # set predecessor (only if not backtracking)
                pred_map[int(nb_idx)] = node_idx
            # otherwise, keep going
            prev_idx = node_idx
            node_idx = int(nb_idx)
        # print(f'[{data_idx}, {nearest}, {next_nearest}],')
        # set in the data map
        data_map[data_idx, 2] = nearest  # adj_idx
        # in some cases next nearest will be NaN
        # this is mostly in situations where it works to leave as NaN - e.g. access off dead-ends...
        data_map[data_idx, 3] = next_nearest  # next_adj_idx

    return data_map


@njit(cache=False)
def aggregate_to_src_idx(netw_src_idx: int,
                         node_data: np.ndarray,
                         edge_data: np.ndarray,
                         node_edge_map: Dict,
                         data_map: np.ndarray,
                         max_dist: float,
                         angular: bool = False):
    # this function is typically called iteratively, so do type checks from parent methods
    netw_x_arr = node_data[:, 0]
    netw_y_arr = node_data[:, 1]
    netw_src_x = netw_x_arr[netw_src_idx]
    netw_src_y = netw_y_arr[netw_src_idx]
    d_x_arr = data_map[:, 0]
    d_y_arr = data_map[:, 1]
    d_assign_nearest = data_map[:, 2]
    d_assign_next_nearest = data_map[:, 3]
    # run the shortest tree dijkstra
    # keep in mind that predecessor map is based on impedance heuristic - which can be different from metres
    # NOTE -> use np.inf for max distance so as to explore all paths
    # In some cases the predecessor nodes will be within reach even if the closest node is not
    # Total distance is checked later
    tree_map, tree_edges = centrality.shortest_path_tree(edge_data,
                                                         node_edge_map,
                                                         netw_src_idx,
                                                         max_dist=max_dist,
                                                         angular=angular)  # turn off checks! This is called iteratively...
    tree_preds = tree_map[:, 1]
    tree_dists = tree_map[:, 2]
    # filter the data by distance
    # in this case, the source x, y is the same as for the networks
    filtered_data = radial_filter(netw_src_x, netw_src_y, d_x_arr, d_y_arr, max_dist)
    # arrays for writing the reachable data points and their distances
    reachable_data = np.full(len(data_map), False)
    reachable_data_dist = np.full(len(data_map), np.inf)
    # iterate the distance trimmed data points
    reachable_idx = np.where(filtered_data)[0]
    for data_idx in reachable_idx:
        # find the primary assigned network index for the data point
        if np.isfinite(d_assign_nearest[data_idx]):
            netw_idx = int(d_assign_nearest[data_idx])
            # if the assigned network node is within the threshold
            if tree_dists[netw_idx] < max_dist:
                # get the distance from the data point to the network node
                d_d = np.hypot(d_x_arr[data_idx] - netw_x_arr[netw_idx],
                               d_y_arr[data_idx] - netw_y_arr[netw_idx])
                # add to the distance assigned for the network node
                dist = tree_dists[netw_idx] + d_d
                # only assign distance if within max distance
                if dist <= max_dist:
                    reachable_data[data_idx] = True
                    reachable_data_dist[data_idx] = dist
        # the next-nearest may offer a closer route depending on the direction the shortest path approaches from
        if np.isfinite(d_assign_next_nearest[data_idx]):
            netw_idx = int(d_assign_next_nearest[data_idx])
            # if the assigned network node is within the threshold
            if tree_dists[netw_idx] < max_dist:
                # get the distance from the data point to the network node
                d_d = np.hypot(d_x_arr[data_idx] - netw_x_arr[netw_idx],
                               d_y_arr[data_idx] - netw_y_arr[netw_idx])
                # add to the distance assigned for the network node
                dist = tree_dists[netw_idx] + d_d
                # only assign distance if within max distance
                # AND only if closer than other direction
                if dist <= max_dist and dist < reachable_data_dist[data_idx]:
                    reachable_data[data_idx] = True
                    reachable_data_dist[data_idx] = dist

    # note that some entries will be nan values if the max distance was exceeded
    return reachable_data, reachable_data_dist, tree_preds


@njit(cache=True)
def local_aggregator(node_data: np.ndarray,
                     edge_data: np.ndarray,
                     node_edge_map: Dict,
                     data_map: np.ndarray,
                     distances: np.ndarray,
                     betas: np.ndarray,
                     landuse_encodings: np.ndarray = np.array([]),
                     qs: np.ndarray = np.array([]),
                     mixed_use_hill_keys: np.ndarray = np.array([]),
                     mixed_use_other_keys: np.ndarray = np.array([]),
                     accessibility_keys: np.ndarray = np.array([]),
                     cl_disparity_wt_matrix: np.ndarray = np.array(np.full((0, 0), np.nan)),
                     numerical_arrays: np.ndarray = np.array(np.full((0, 0), np.nan)),
                     angular: bool = False,
                     suppress_progress: bool = False) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray,
                                                               np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    '''
    NODE MAP:
    0 - x
    1 - y
    2 - live
    EDGE MAP:
    0 - start node
    1 - end node
    2 - length in metres
    3 - sum of angular travel along length
    4 - impedance factor
    5 - in bearing
    6 - out bearing
    DATA MAP:
    0 - x
    1 - y
    2 - assigned network index - nearest
    3 - assigned network index - next-nearest
    '''
    checks.check_network_maps(node_data, edge_data, node_edge_map)
    checks.check_data_map(data_map, check_assigned=True)  # raises ValueError data points are not assigned to a network
    checks.check_distances_and_betas(distances, betas)

    # check landuse encodings
    compute_landuses = False
    if len(landuse_encodings) == 0:
        if len(mixed_use_hill_keys) != 0 or len(mixed_use_other_keys) != 0 or len(accessibility_keys) != 0:
            raise ValueError('Mixed use metrics or land-use accessibilities require an array of landuse labels.')
    elif len(landuse_encodings) != len(data_map):
        raise ValueError('The number of landuse encodings does not match the number of data points.')
    else:
        checks.check_categorical_data(landuse_encodings)

    # catch completely missing metrics
    if len(mixed_use_hill_keys) == 0 and len(mixed_use_other_keys) == 0 and len(accessibility_keys) == 0:
        if len(numerical_arrays) == 0:
            raise ValueError(
                'No metrics specified, please specify at least one metric to compute.')
    else:
        compute_landuses = True

    # catch missing qs
    if len(mixed_use_hill_keys) != 0 and len(qs) == 0:
        raise ValueError('Hill diversity measures require that at least one value of q is specified.')

    # negative qs caught by hill diversity methods

    # check various problematic key combinations
    if len(mixed_use_hill_keys) != 0:
        if (mixed_use_hill_keys.min() < 0 or mixed_use_hill_keys.max() > 3):
            raise ValueError('Mixed-use "hill" keys out of range of 0:4.')

    if len(mixed_use_other_keys) != 0:
        if (mixed_use_other_keys.min() < 0 or mixed_use_other_keys.max() > 2):
            raise ValueError('Mixed-use "other" keys out of range of 0:3.')

    if len(accessibility_keys) != 0:
        max_ac_key = landuse_encodings.max()
        if (accessibility_keys.min() < 0 or accessibility_keys.max() > max_ac_key):
            raise ValueError('Negative or out of range accessibility key encountered. Keys must match class encodings.')

    for i in range(len(mixed_use_hill_keys)):
        for j in range(len(mixed_use_hill_keys)):
            if j > i:
                i_key = mixed_use_hill_keys[i]
                j_key = mixed_use_hill_keys[j]
                if i_key == j_key:
                    raise ValueError('Duplicate mixed-use "hill" key.')

    for i in range(len(mixed_use_other_keys)):
        for j in range(len(mixed_use_other_keys)):
            if j > i:
                i_key = mixed_use_other_keys[i]
                j_key = mixed_use_other_keys[j]
                if i_key == j_key:
                    raise ValueError('Duplicate mixed-use "other" key.')

    for i in range(len(accessibility_keys)):
        for j in range(len(accessibility_keys)):
            if j > i:
                i_key = accessibility_keys[i]
                j_key = accessibility_keys[j]
                if i_key == j_key:
                    raise ValueError('Duplicate accessibility key.')

    def disp_check(disp_matrix):
        # the length of the disparity matrix vis-a-vis unique landuses is tested in underlying diversity functions
        if disp_matrix.ndim != 2 or disp_matrix.shape[0] != disp_matrix.shape[1]:
            raise ValueError('The disparity matrix must be a square NxN matrix.')
        if len(disp_matrix) == 0:
            raise ValueError('Hill disparity and Rao pairwise measures requires a class disparity weights matrix.')

    # check that missing or malformed disparity weights matrices are caught
    for k in mixed_use_hill_keys:
        if k == 3:  # hill disparity
            disp_check(cl_disparity_wt_matrix)
    for k in mixed_use_other_keys:
        if k == 2:  # raos pairwise
            disp_check(cl_disparity_wt_matrix)

    compute_numerical = False
    # when passing an empty 2d array to numba, use: np.array(np.full((0, 0), np.nan))
    if len(numerical_arrays) != 0:
        compute_numerical = True
        if numerical_arrays.shape[1] != len(data_map):
            raise ValueError('The length of the numerical data arrays do not match the length of the data map.')
        checks.check_numerical_data(numerical_arrays)

    # establish variables
    netw_n = len(node_data)
    d_n = len(distances)
    q_n = len(qs)
    n_n = len(numerical_arrays)
    global_max_dist = distances.max()
    netw_nodes_live = node_data[:, 2]

    # setup data structures
    # hill mixed uses are structured separately to take values of q into account
    mixed_use_hill_data = np.full((4, q_n, d_n, netw_n), np.nan)  # 4 dim
    mixed_use_other_data = np.full((3, d_n, netw_n), np.nan)  # 3 dim

    accessibility_data = np.full((len(accessibility_keys), d_n, netw_n), 0.0)
    accessibility_data_wt = np.full((len(accessibility_keys), d_n, netw_n), 0.0)

    # stats
    stats_sum = np.full((n_n, d_n, netw_n), np.nan)
    stats_sum_wt = np.full((n_n, d_n, netw_n), np.nan)

    stats_mean = np.full((n_n, d_n, netw_n), np.nan)
    stats_mean_wt = np.full((n_n, d_n, netw_n), np.nan)

    stats_count = np.full((n_n, d_n, netw_n), np.nan)  # use np.nan instead of 0 to avoid division by zero issues
    stats_count_wt = np.full((n_n, d_n, netw_n), np.nan)

    stats_variance = np.full((n_n, d_n, netw_n), np.nan)
    stats_variance_wt = np.full((n_n, d_n, netw_n), np.nan)

    stats_max = np.full((n_n, d_n, netw_n), np.nan)
    stats_min = np.full((n_n, d_n, netw_n), np.nan)

    # iterate through each vert and aggregate
    steps = int(netw_n / 10000)
    for netw_src_idx in range(netw_n):
        if not suppress_progress:
            checks.progress_bar(netw_src_idx, netw_n, steps)
        # only compute for live nodes
        if not netw_nodes_live[netw_src_idx]:
            continue
        # generate the reachable classes and their respective distances
        # these are non-unique - i.e. simply the class of each data point within the maximum distance
        # the aggregate_to_src_idx method will choose the closer direction of approach to a data point
        # from the nearest or next-nearest network node (calculated once globally, prior to local_landuses method)
        reachable_data, reachable_data_dist, tree_preds = aggregate_to_src_idx(netw_src_idx,
                                                                               node_data,
                                                                               edge_data,
                                                                               node_edge_map,
                                                                               data_map,
                                                                               global_max_dist,
                                                                               angular)
        # LANDUSES
        if compute_landuses:
            mu_max_unique_cl = int(landuse_encodings.max() + 1)
            # counts of each class type (array length per max unique classes - not just those within max distance)
            classes_counts = np.full((d_n, mu_max_unique_cl), 0)
            # nearest of each class type (likewise)
            classes_nearest = np.full((d_n, mu_max_unique_cl), np.inf)
            # iterate the reachable indices and related distances
            for data_idx, (reachable, data_dist) in enumerate(zip(reachable_data, reachable_data_dist)):
                if not reachable:
                    continue
                # get the class category in integer form
                # all class codes were encoded to sequential integers - these correspond to the array indices
                cl_code = int(landuse_encodings[int(data_idx)])
                # iterate the distance dimensions
                for d_idx, (d, b) in enumerate(zip(distances, betas)):
                    # increment class counts at respective distances if the distance is less than current d
                    if data_dist <= d:
                        classes_counts[d_idx, cl_code] += 1
                        # if distance is nearer, update the nearest distance array too
                        if data_dist < classes_nearest[d_idx, cl_code]:
                            classes_nearest[d_idx, cl_code] = data_dist
                        # if within distance, and if in accessibility keys, then aggregate accessibility too
                        for ac_idx, ac_code in enumerate(accessibility_keys):
                            if ac_code == cl_code:
                                accessibility_data[ac_idx, d_idx, netw_src_idx] += 1
                                accessibility_data_wt[ac_idx, d_idx, netw_src_idx] += np.exp(b * data_dist)
                                # if a match was found, then no need to check others
                                break
            # mixed uses can be calculated now that the local class counts are aggregated
            # iterate the distances and betas
            for d_idx, b in enumerate(betas):
                cl_counts = classes_counts[d_idx]
                cl_nearest = classes_nearest[d_idx]
                # mu keys determine which metrics to compute
                # don't confuse with indices
                # previously used dynamic indices in data structures - but obtuse if irregularly ordered keys
                for mu_hill_key in mixed_use_hill_keys:
                    for q_idx, q_key in enumerate(qs):
                        if mu_hill_key == 0:
                            mixed_use_hill_data[0, q_idx, d_idx, netw_src_idx] = \
                                diversity.hill_diversity(cl_counts, q_key)
                        elif mu_hill_key == 1:
                            mixed_use_hill_data[1, q_idx, d_idx, netw_src_idx] = \
                                diversity.hill_diversity_branch_distance_wt(cl_counts, cl_nearest, q=q_key, beta=b)
                        elif mu_hill_key == 2:
                            mixed_use_hill_data[2, q_idx, d_idx, netw_src_idx] = \
                                diversity.hill_diversity_pairwise_distance_wt(cl_counts, cl_nearest, q=q_key, beta=b)
                        # land-use classification disparity hill diversity
                        # the wt matrix can be used without mapping because cl_counts is based on all classes
                        # regardless of whether they are reachable
                        elif mu_hill_key == 3:
                            mixed_use_hill_data[3, q_idx, d_idx, netw_src_idx] = \
                                diversity.hill_diversity_pairwise_matrix_wt(cl_counts,
                                                                            wt_matrix=cl_disparity_wt_matrix,
                                                                            q=q_key)
                for mu_other_key in mixed_use_other_keys:
                    if mu_other_key == 0:
                        mixed_use_other_data[0, d_idx, netw_src_idx] = \
                            diversity.shannon_diversity(cl_counts)
                    elif mu_other_key == 1:
                        mixed_use_other_data[1, d_idx, netw_src_idx] = \
                            diversity.gini_simpson_diversity(cl_counts)
                    elif mu_other_key == 2:
                        mixed_use_other_data[2, d_idx, netw_src_idx] = \
                            diversity.raos_quadratic_diversity(cl_counts, wt_matrix=cl_disparity_wt_matrix)
        # IDW
        # the order of the loops matters because the nested aggregations happen per distance per numerical array
        if compute_numerical:
            # iterate the reachable indices and related distances
            for data_idx, (reachable, data_dist) in enumerate(zip(reachable_data, reachable_data_dist)):
                # some indices will be NaN if beyond max threshold distance - so check for infinity
                # this happens when within radial max distance, but beyond network max distance
                if not reachable:
                    continue
                # iterate the numerical arrays dimension
                for num_idx in range(n_n):
                    # some values will be NaN
                    num = numerical_arrays[num_idx, int(data_idx)]
                    if np.isnan(num):
                        continue
                    # iterate the distance dimensions
                    for d_idx, (d, b) in enumerate(zip(distances, betas)):
                        # increment mean aggregations at respective distances if the distance is less than current d
                        if data_dist <= d:
                            # aggregate
                            if np.isnan(stats_sum[num_idx, d_idx, netw_src_idx]):
                                stats_sum[num_idx, d_idx, netw_src_idx] = num
                                stats_count[num_idx, d_idx, netw_src_idx] = 1
                                stats_sum_wt[num_idx, d_idx, netw_src_idx] = num * np.exp(data_dist * b)
                                stats_count_wt[num_idx, d_idx, netw_src_idx] = np.exp(data_dist * b)
                            else:
                                stats_sum[num_idx, d_idx, netw_src_idx] += num
                                stats_count[num_idx, d_idx, netw_src_idx] += 1
                                stats_sum_wt[num_idx, d_idx, netw_src_idx] += num * np.exp(data_dist * b)
                                stats_count_wt[num_idx, d_idx, netw_src_idx] += np.exp(data_dist * b)

                            if np.isnan(stats_max[num_idx, d_idx, netw_src_idx]):
                                stats_max[num_idx, d_idx, netw_src_idx] = num
                            elif num > stats_max[num_idx, d_idx, netw_src_idx]:
                                stats_max[num_idx, d_idx, netw_src_idx] = num

                            if np.isnan(stats_min[num_idx, d_idx, netw_src_idx]):
                                stats_min[num_idx, d_idx, netw_src_idx] = num
                            elif num < stats_min[num_idx, d_idx, netw_src_idx]:
                                stats_min[num_idx, d_idx, netw_src_idx] = num
            # finalise mean calculations - this is happening for a single netw_src_idx, so fairly fast
            for num_idx in range(n_n):
                for d_idx in range(d_n):
                    stats_mean[num_idx, d_idx, netw_src_idx] = \
                        stats_sum[num_idx, d_idx, netw_src_idx] / stats_count[num_idx, d_idx, netw_src_idx]
                    stats_mean_wt[num_idx, d_idx, netw_src_idx] = \
                        stats_sum_wt[num_idx, d_idx, netw_src_idx] / stats_count_wt[num_idx, d_idx, netw_src_idx]
            # calculate variances - counts are already computed per above
            # weighted version is IDW by division through equivalently weighted counts above
            # iterate the reachable indices and related distances
            for data_idx, (reachable, data_dist) in enumerate(zip(reachable_data, reachable_data_dist)):
                # some indices will be NaN if beyond max threshold distance - so check for infinity
                # this happens when within radial max distance, but beyond network max distance
                if not reachable:
                    continue
                # iterate the numerical arrays dimension
                for num_idx in range(n_n):
                    # some values will be NaN
                    num = numerical_arrays[num_idx, int(data_idx)]
                    if np.isnan(num):
                        continue
                    # iterate the distance dimensions
                    for d_idx, (d, b) in enumerate(zip(distances, betas)):
                        # increment variance aggregations at respective distances if the distance is less than current d
                        if data_dist <= d:
                            # aggregate
                            if np.isnan(stats_variance[num_idx, d_idx, netw_src_idx]):
                                stats_variance[num_idx, d_idx, netw_src_idx] = \
                                    np.square(num - stats_mean[num_idx, d_idx, netw_src_idx])
                                stats_variance_wt[num_idx, d_idx, netw_src_idx] = \
                                    np.square(num - stats_mean_wt[num_idx, d_idx, netw_src_idx]) * np.exp(data_dist * b)
                            else:
                                stats_variance[num_idx, d_idx, netw_src_idx] += \
                                    np.square(num - stats_mean[num_idx, d_idx, netw_src_idx])
                                stats_variance_wt[num_idx, d_idx, netw_src_idx] += \
                                    np.square(num - stats_mean_wt[num_idx, d_idx, netw_src_idx]) * np.exp(data_dist * b)
            # finalise variance calculations
            for num_idx in range(n_n):
                for d_idx in range(d_n):
                    stats_variance[num_idx, d_idx, netw_src_idx] = \
                        stats_variance[num_idx, d_idx, netw_src_idx] / stats_count[num_idx, d_idx, netw_src_idx]
                    stats_variance_wt[num_idx, d_idx, netw_src_idx] = \
                        stats_variance_wt[num_idx, d_idx, netw_src_idx] / stats_count_wt[num_idx, d_idx, netw_src_idx]
    # send the data back in the same types and same order as the original keys - convert to int for indexing
    mu_hill_k_int = np.full(len(mixed_use_hill_keys), 0)
    for i, k in enumerate(mixed_use_hill_keys):
        mu_hill_k_int[i] = k
    mu_other_k_int = np.full(len(mixed_use_other_keys), 0)
    for i, k in enumerate(mixed_use_other_keys):
        mu_other_k_int[i] = k

    return mixed_use_hill_data[mu_hill_k_int], \
           mixed_use_other_data[mu_other_k_int], \
           accessibility_data, accessibility_data_wt, \
           stats_sum, stats_sum_wt, \
           stats_mean, stats_mean_wt, \
           stats_variance, stats_variance_wt, \
           stats_max, stats_min


@njit(cache=True)
def singly_constrained(node_data: np.ndarray,
                       edge_data: np.ndarray,
                       node_edge_map: Dict,
                       distances: np.ndarray,
                       betas: np.ndarray,
                       i_data_map: np.ndarray,
                       j_data_map: np.ndarray,
                       i_weights: np.ndarray,
                       j_weights: np.ndarray,
                       angular: bool = False,
                       suppress_progress: bool = False) -> Tuple[np.ndarray, np.ndarray]:
    '''
    - Calculates trips from i to j and returns the assigned trips and network assigned flows for j nodes
    #TODO: consider enhanced numerical checks for single vs. multi dimensional numerical data

    - Keeping separate from local aggregator because singly-constrained origin / destination models computed separately
    - Requires two iters, one to gather all k-nodes to per j node, then another to get the ratio of j / k attractiveness
    - Assigns j -> k trips over the network as part of second iter
    NODE MAP:
    0 - x
    1 - y
    2 - live
    EDGE MAP:
    0 - start node
    1 - end node
    2 - length in metres
    3 - sum of angular travel along length
    4 - impedance factor
    5 - entry bearing
    6 - exit bearing
    DATA MAP:
    0 - x
    1 - y
    2 - assigned network index - nearest
    3 - assigned network index - next-nearest
    '''
    checks.check_network_maps(node_data, edge_data, node_edge_map)
    checks.check_distances_and_betas(distances, betas)
    checks.check_data_map(i_data_map, check_assigned=True)
    checks.check_data_map(j_data_map, check_assigned=True)

    if len(i_weights) != len(i_data_map):
        raise ValueError('The i_weights array must be the same length as the i_data_map.')

    if len(j_weights) != len(j_data_map):
        raise ValueError('The j_weights array must be the same length as the j_data_map.')

    # establish variables
    netw_n = len(node_data)
    d_n = len(distances)
    global_max_dist = np.max(distances)
    netw_flows = np.full((d_n, netw_n), 0.0)

    i_n = len(i_data_map)
    k_agg = np.full((d_n, i_n), 0.0)

    j_n = len(j_data_map)
    j_assigned = np.full((d_n, j_n), 0.0)

    # iterate all i nodes
    # filter all reachable nodes k and aggregate k attractiveness * negative exponential of distance
    steps = int(i_n / 10000)
    for i_idx in range(i_n):
        if not suppress_progress:
            checks.progress_bar(i_idx, i_n, steps)
        # get the nearest node
        i_assigned_netw_idx = int(i_data_map[i_idx, 2])
        # calculate the base distance from the data point to the nearest assigned node
        i_x, i_y = i_data_map[i_idx, :2]
        n_x, n_y = node_data[i_assigned_netw_idx, :2]
        i_door_dist = np.hypot(i_x - n_x, i_y - n_y)

        # find the reachable j data points and their respective points from the closest node
        reachable_j, reachable_j_dist, tree_preds = aggregate_to_src_idx(i_assigned_netw_idx,
                                                                         node_data,
                                                                         edge_data,
                                                                         node_edge_map,
                                                                         j_data_map,
                                                                         global_max_dist,
                                                                         angular)

        # aggregate the weighted j (all k) nodes
        # iterate the reachable indices and related distances
        for j_idx, (j_reachable, j_dist) in enumerate(zip(reachable_j, reachable_j_dist)):
            if not j_reachable:
                continue
            # iterate the distance dimensions
            for d_idx, (d, b) in enumerate(zip(distances, betas)):
                total_dist = j_dist + i_door_dist
                # increment weighted k aggregations at respective distances if the distance is less than current d
                if total_dist <= d:
                    k_agg[d_idx, i_idx] += j_weights[j_idx] * np.exp(total_dist * b)

    # this is the second step
    # this time, filter all reachable j vertices and aggregate the proportion of flow from i to j
    # this is done by dividing i-j flow through i-k_agg flow from previous step
    steps = int(i_n / 10000)
    for i_idx in range(i_n):
        if not suppress_progress:
            checks.progress_bar(i_idx, i_n, steps)

        # get the nearest node
        i_assigned_netw_idx = int(i_data_map[i_idx, 2])
        # calculate the base distance from the data point to the nearest assigned node
        i_x, i_y = i_data_map[i_idx, :2]
        n_x, n_y = node_data[i_assigned_netw_idx, :2]
        i_door_dist = np.hypot(i_x - n_x, i_y - n_y)

        # find the reachable j data points and their respective points from the closest node
        reachable_j, reachable_j_dist, tree_preds = aggregate_to_src_idx(i_assigned_netw_idx,
                                                                         node_data,
                                                                         edge_data,
                                                                         node_edge_map,
                                                                         j_data_map,
                                                                         global_max_dist,
                                                                         angular)

        # aggregate j divided through all k nodes
        # iterate the reachable indices and related distances
        for j_idx, (j_reachable, j_dist) in enumerate(zip(reachable_j, reachable_j_dist)):
            if not j_reachable:
                continue
            # iterate the distance dimensions
            for d_idx, (d, b) in enumerate(zip(distances, betas)):
                total_dist = j_dist + i_door_dist
                # if the distance is less than current d
                if total_dist <= d:
                    # aggregate all flows from reachable j's to i_idx
                    # divide through respective i-k_agg sums
                    # catch division by zero:
                    if k_agg[d_idx, i_idx] == 0:
                        assigned = 0
                    else:
                        assigned = i_weights[i_idx] * j_weights[j_idx] * np.exp(total_dist * b) / k_agg[d_idx,
                                                                                                        i_idx]
                    j_assigned[d_idx, j_idx] += assigned
                    # assign trips to network
                    if assigned != 0:
                        # get the j assigned node
                        j_assigned_netw_idx = int(j_data_map[j_idx, 2])
                        # in this case start and end nodes are counted...!
                        netw_flows[d_idx, j_assigned_netw_idx] += assigned
                        # skip if same start / end node
                        if j_assigned_netw_idx == i_assigned_netw_idx:
                            continue
                        # aggregate to the network
                        inter_idx = np.int(tree_preds[j_assigned_netw_idx])
                        while True:
                            # end nodes counted, so place above break
                            netw_flows[d_idx, inter_idx] += assigned
                            # break out of while loop if the intermediary has reached the source node
                            if inter_idx == i_assigned_netw_idx:
                                break
                            # follow the chain
                            inter_idx = np.int(tree_preds[inter_idx])

    return j_assigned, netw_flows
